# File from the Spring benchmark
# https://spring-benchmark.org/

import struct
import numpy as np
import png
import re
import sys
import csv
from PIL import Image
import h5py
import cv2 as cv


FLO_TAG_FLOAT = (
    202021.25  # first 4 bytes in flo file; check for this when READING the file
)
FLO_TAG_STRING = "PIEH"  # first 4 bytes in flo file; use this when WRITING the file
FLO_UNKNOWN_FLOW_THRESH = 1e9  # flo format threshold for unknown values
FLO_UNKNOWN_FLOW = 1e10  # value to use to represent unknown flow in flo file format


def readFlowFile(filepath):
    """read flow files in several formats. The resulting flow has shape height x width x 2.
    For positions where there is no groundtruth available, the flow is set to np.nan.
    Supports flo (Sintel), png (KITTI), npy (numpy), pfm (FlyingThings3D) and flo5 (Spring) file format.
    filepath: path to the flow file
    returns: flow with shape height x width x 2
    """
    if filepath.endswith(".flo"):
        return readFloFlow(filepath)
    elif filepath.endswith(".png"):
        return readPngFlow(filepath)
    elif filepath.endswith(".npy"):
        return readNpyFlow(filepath)
    elif filepath.endswith(".pfm"):
        return readPfmFlow(filepath)
    elif filepath.endswith(".flo5"):
        return readFlo5Flow(filepath)
    else:
        raise ValueError(f"readFlowFile: Unknown file format for {filepath}")


def writeFlowFile(flow, filepath):
    """write optical flow to file. Supports flo (Sintel), png (KITTI) and npy (numpy) file format.
    flow: optical flow with shape height x width x 2. Invalid values should be represented as np.nan
    filepath: file path where to write the flow
    """
    if not filepath:
        raise ValueError("writeFlowFile: empty filepath")

    if len(flow.shape) != 3 or flow.shape[2] != 2:
        raise IOError(
            f"writeFlowFile {filepath}: expected shape height x width x 2 but received {flow.shape}"
        )

    if flow.shape[0] > flow.shape[1]:
        print(
            f"write flo file {filepath}: Warning: Are you writing an upright image? Expected shape height x width x 2, got {flow.shape}"
        )

    if filepath.endswith(".flo"):
        return writeFloFlow(flow, filepath)
    elif filepath.endswith(".png"):
        return writePngFlow(flow, filepath)
    elif filepath.endswith(".npy"):
        return writeNpyFile(flow, filepath)
    elif filepath.endswith(".flo5"):
        return writeFlo5File(flow, filepath)
    else:
        raise ValueError(f"writeFlowFile: Unknown file format for {filepath}")


def readFloFlow(filepath):
    """read optical flow from file stored in .flo file format as used in the Sintel dataset (Butler et al., 2012)
    filepath: path to file where to read from
    returns: flow as a numpy array with shape height x width x 2
    ---
    ".flo" file format used for optical flow evaluation
    Stores 2-band float image for horizontal (u) and vertical (v) flow components.
    Floats are stored in little-endian order.
    A flow value is considered "unknown" if either |u| or |v| is greater than 1e9.
    bytes  contents
    0-3     tag: "PIEH" in ASCII, which in little endian happens to be the float 202021.25
            (just a sanity check that floats are represented correctly)
    4-7     width as an integer
    8-11    height as an integer
    12-end  data (width*height*2*4 bytes total)
            the float values for u and v, interleaved, in row order, i.e.,
            u[row0,col0], v[row0,col0], u[row0,col1], v[row0,col1], ...
    """
    if filepath is None:
        raise IOError("read flo file: empty filename")

    if not filepath.endswith(".flo"):
        raise IOError(f"read flo file ({filepath}): extension .flo expected")

    with open(filepath, "rb") as stream:
        tag = struct.unpack("f", stream.read(4))[0]
        width = struct.unpack("i", stream.read(4))[0]
        height = struct.unpack("i", stream.read(4))[0]

        if tag != FLO_TAG_FLOAT:  # simple test for correct endian-ness
            raise IOError(
                f"read flo file({filepath}): wrong tag (possibly due to big-endian machine?)"
            )

        # another sanity check to see that integers were read correctly (99999 should do the trick...)
        if width < 1 or width > 99999:
            raise IOError(f"read flo file({filepath}): illegal width {width}")

        if height < 1 or height > 99999:
            raise IOError(f"read flo file({filepath}): illegal height {height}")

        nBands = 2
        flow = []

        n = nBands * width
        for _ in range(height):
            data = stream.read(n * 4)
            if data is None:
                raise IOError(f"read flo file({filepath}): file is too short")
            data = np.asarray(struct.unpack(f"{n}f", data))
            data = data.reshape((width, nBands))
            flow.append(data)

        if stream.read(1) != b"":
            raise IOError(f"read flo file({filepath}): file is too long")

        flow = np.asarray(flow)
        # unknown values are set to nan
        flow[np.abs(flow) > FLO_UNKNOWN_FLOW_THRESH] = np.nan

        return flow


def writeFloFlow(flow, filepath):
    """
    write optical flow in .flo format to file as used in the Sintel dataset (Butler et al., 2012)
    flow: optical flow with shape height x width x 2
    filepath: optical flow file path to be saved
    ---
    ".flo" file format used for optical flow evaluation
    Stores 2-band float image for horizontal (u) and vertical (v) flow components.
    Floats are stored in little-endian order.
    A flow value is considered "unknown" if either |u| or |v| is greater than 1e9.
    bytes  contents
    0-3     tag: "PIEH" in ASCII, which in little endian happens to be the float 202021.25
            (just a sanity check that floats are represented correctly)
    4-7     width as an integer
    8-11    height as an integer
    12-end  data (width*height*2*4 bytes total)
            the float values for u and v, interleaved, in row order, i.e.,
            u[row0,col0], v[row0,col0], u[row0,col1], v[row0,col1], ...
    """

    height, width, nBands = flow.shape

    with open(filepath, "wb") as f:
        if f is None:
            raise IOError(f"write flo file {filepath}: file could not be opened")

        # write header
        result = f.write(FLO_TAG_STRING.encode("ascii"))
        result += f.write(struct.pack("i", width))
        result += f.write(struct.pack("i", height))
        if result != 12:
            raise IOError(f"write flo file {filepath}: problem writing header")

        # write content
        n = nBands * width
        for i in range(height):
            data = flow[i, :, :].flatten()
            data[np.isnan(data)] = FLO_UNKNOWN_FLOW
            result = f.write(struct.pack(f"{n}f", *data))
            if result != n * 4:
                raise IOError(f"write flo file {filepath}: problem writing row {i}")


def readPngFlow(filepath):
    """read optical flow from file stored in png file format as used in the KITTI 12 (Geiger et al., 2012) and KITTI 15 (Menze et al., 2015) dataset.
    filepath: path to file where to read from
    returns: flow as a numpy array with shape height x width x 2. Invalid values are represented as np.nan
    """
    # adapted from https://github.com/liruoteng/OpticalFlowToolkit
    flow_object = png.Reader(filename=filepath)
    flow_direct = flow_object.asDirect()
    flow_data = list(flow_direct[2])
    (w, h) = flow_direct[3]["size"]
    flow = np.zeros((h, w, 3), dtype=np.float64)
    for i in range(len(flow_data)):
        flow[i, :, 0] = flow_data[i][0::3]
        flow[i, :, 1] = flow_data[i][1::3]
        flow[i, :, 2] = flow_data[i][2::3]

    invalid_idx = flow[:, :, 2] == 0
    flow[:, :, 0:2] = (flow[:, :, 0:2] - 2**15) / 64.0
    flow[invalid_idx, 0] = np.nan
    flow[invalid_idx, 1] = np.nan
    return flow[:, :, :2]


def writePngFlow(flow, filename):
    """write optical flow to file png file format as used in the KITTI 12 (Geiger et al., 2012) and KITTI 15 (Menze et al., 2015) dataset.
    flow: optical flow in shape height x width x 2, invalid values should be represented as np.nan
    filepath: path to file where to write to
    """
    flow = 64.0 * flow + 2**15
    width = flow.shape[1]
    height = flow.shape[0]
    valid_map = np.ones([flow.shape[0], flow.shape[1], 1])
    valid_map[np.isnan(flow[:, :, 0]) | np.isnan(flow[:, :, 1])] = 0
    flow = np.nan_to_num(flow)
    flow = np.concatenate([flow, valid_map], axis=-1)
    flow = np.clip(flow, 0, 2**16 - 1)
    flow = flow.astype(np.uint16)
    flow = np.reshape(flow, (-1, width * 3))
    with open(filename, "wb") as f:
        writer = png.Writer(width=width, height=height, bitdepth=16, greyscale=False)
        writer.write(f, flow)


def readNpyFlow(filepath):
    """read numpy array from file.
    filepath: file to read from
    returns: numpy array
    """
    return np.load(filepath)


def writeNpyFile(arr, filepath):
    """write numpy array to file.
    arr: numpy array to write
    filepath: file to write to
    """
    np.save(filepath, arr)


def writeFlo5File(flow, filename):
    with h5py.File(filename, "w") as f:
        f.create_dataset("flow", data=flow, compression="gzip", compression_opts=5)


def readFlo5Flow(filename):
    with h5py.File(filename, "r") as f:
        if "flow" not in f.keys():
            raise IOError(
                f"File {filename} does not have a 'flow' key. Is this a valid flo5 file?"
            )
        return f["flow"][()]


def readPfmFlow(filepath):
    """read optical flow from file stored in pfm file format as used in the FlyingThings3D (Mayer et al., 2016) dataset.
    filepath: path to file where to read from
    returns: flow as a numpy array with shape height x width x 2.
    """
    flow = readPfmFile(filepath)
    if len(flow.shape) != 3:
        raise IOError(
            f"read pfm flow: PFM file has wrong shape (assumed to be w x h x 3): {flow.shape}"
        )
    if flow.shape[2] != 3:
        raise IOError(
            f"read pfm flow: PFM file has wrong shape (assumed to be w x h x 3): {flow.shape}"
        )
    # remove third channel -> is all zeros
    return flow[:, :, :2]


def readPfmFile(filepath):
    """
    adapted from https://lmb.informatik.uni-freiburg.de/resources/datasets/SceneFlowDatasets.en.html
    """
    file = open(filepath, "rb")

    color = None
    width = None
    height = None
    scale = None
    endian = None

    header = file.readline().rstrip()
    if header.decode("ascii") == "PF":
        color = True
    elif header.decode("ascii") == "Pf":
        color = False
    else:
        raise Exception("Not a PFM file.")

    dim_match = re.match(r"^(\d+)\s(\d+)\s$", file.readline().decode("ascii"))
    if dim_match:
        width, height = list(map(int, dim_match.groups()))
    else:
        raise Exception("Malformed PFM header.")

    scale = float(file.readline().decode("ascii").rstrip())
    if scale < 0:  # little-endian
        endian = "<"
        scale = -scale
    else:
        endian = ">"  # big-endian

    data = np.fromfile(file, endian + "f")
    shape = (height, width, 3) if color else (height, width)

    data = np.reshape(data, shape)
    data = np.flipud(data)
    return data  # , scale


def writePfmFile(image, filepath):
    """
    adapted from https://lmb.informatik.uni-freiburg.de/resources/datasets/SceneFlowDatasets.en.html
    """
    scale = 1
    file = open(filepath, "wb")

    color = None

    if image.dtype.name != "float32":
        raise Exception("Image dtype must be float32.")

    image = np.flipud(image)

    if len(image.shape) == 3 and image.shape[2] == 3:  # color image
        color = True
    elif (
        len(image.shape) == 2 or len(image.shape) == 3 and image.shape[2] == 1
    ):  # greyscale
        color = False
    else:
        raise Exception("Image must have H x W x 3, H x W x 1 or H x W dimensions.")

    file.write("PF\n" if color else "Pf\n".encode())
    file.write("%d %d\n".encode() % (image.shape[1], image.shape[0]))

    endian = image.dtype.byteorder

    if endian == "<" or endian == "=" and sys.byteorder == "little":
        scale = -scale

    file.write("%f\n".encode() % scale)

    image.tofile(file)


def readDispFile(filepath):
    """read disparity (or disparity change) from file. The resulting numpy array has shape height x width.
    For positions where there is no groundtruth available, the value is set to np.nan.
    Supports png (KITTI), npy (numpy) and pfm (FlyingThings3D) file format.
    filepath: path to the flow file
    returns: disparity with shape height x width
    """
    if filepath.endswith(".png"):
        return readPngDisp(filepath)
    elif filepath.endswith(".npy"):
        return readNpyFlow(filepath)
    elif filepath.endswith(".pfm"):
        return readPfmDisp(filepath)
    elif filepath.endswith(".dsp5"):
        return readDsp5Disp(filepath)
    else:
        raise ValueError(f"readDispFile: Unknown file format for {filepath}")


def readPngDisp(filepath):
    """read disparity from file stored in png file format as used in the KITTI 12 (Geiger et al., 2012) and KITTI 15 (Menze et al., 2015) dataset.
    filepath: path to file where to read from
    returns: disparity as a numpy array with shape height x width. Invalid values are represented as np.nan
    """
    # adapted from https://github.com/liruoteng/OpticalFlowToolkit
    image_object = png.Reader(filename=filepath)
    image_direct = image_object.asDirect()
    image_data = list(image_direct[2])
    (w, h) = image_direct[3]["size"]
    channel = len(image_data[0]) // w
    if channel != 1:
        raise IOError("read png disp: assumed channels to be 1!")
    disp = np.zeros((h, w), dtype=np.float64)
    for i in range(len(image_data)):
        disp[i, :] = image_data[i][:]
    disp[disp == 0] = np.nan
    return disp[:, :] / 256.0


def readPfmDisp(filepath):
    """read disparity or disparity change from file stored in pfm file format as used in the FlyingThings3D (Mayer et al., 2016) dataset.
    filepath: path to file where to read from
    returns: disparity as a numpy array with shape height x width. Invalid values are represented as np.nan
    """
    disp = readPfmFile(filepath)
    if len(disp.shape) != 2:
        raise IOError(
            f"read pfm disp: PFM file has wrong shape (assumed to be w x h): {disp.shape}"
        )
    return disp


def writePngDisp(disp, filepath):
    """write disparity to png file format as used in the KITTI 12 (Geiger et al., 2012) and KITTI 15 (Menze et al., 2015) dataset.
    disp: disparity in shape height x width, invalid values should be represented as np.nan
    filepath: path to file where to write to
    """
    disp = 256 * disp
    width = disp.shape[1]
    height = disp.shape[0]
    disp = np.clip(disp, 0, 2**16 - 1)
    disp = np.nan_to_num(disp).astype(np.uint16)
    disp = np.reshape(disp, (-1, width))
    with open(filepath, "wb") as f:
        writer = png.Writer(width=width, height=height, bitdepth=16, greyscale=True)
        writer.write(f, disp)


def writeDsp5File(disp, filename):
    with h5py.File(filename, "w") as f:
        f.create_dataset("disparity", data=disp, compression="gzip", compression_opts=5)


def readDsp5Disp(filename):
    with h5py.File(filename, "r") as f:
        if "disparity" not in f.keys():
            raise IOError(
                f"File {filename} does not have a 'disparity' key. Is this a valid dsp5 file?"
            )
        return f["disparity"][()]


def writeDispFile(disp, filepath):
    """write disparity to file. Supports png (KITTI) and npy (numpy) file format.
    disp: disparity with shape height x width. Invalid values should be represented as np.nan
    filepath: file path where to write the flow
    """
    if not filepath:
        raise ValueError("writeDispFile: empty filepath")

    if len(disp.shape) != 2:
        raise IOError(
            f"writeDispFile {filepath}: expected shape height x width but received {disp.shape}"
        )

    if disp.shape[0] > disp.shape[1]:
        print(
            f"writeDispFile {filepath}: Warning: Are you writing an upright image? Expected shape height x width, got {disp.shape}"
        )

    if filepath.endswith(".png"):
        writePngDisp(disp, filepath)
    elif filepath.endswith(".npy"):
        writeNpyFile(disp, filepath)
    elif filepath.endswith(".dsp5"):
        writeDsp5File(disp, filepath)
    elif filepath.endswith(".pfm"):
        writePfmFile(disp, filepath)


def readKITTIObjMap(filepath):
    assert filepath.endswith(".png")
    return np.asarray(Image.open(filepath)) > 0


def readKITTIIntrinsics(filepath, image=2):
    assert filepath.endswith(".txt")

    with open(filepath) as f:
        reader = csv.reader(f, delimiter=" ")
        for row in reader:
            if row[0] == f"K_{image:02d}:":
                K = np.array(row[1:], dtype=np.float32).reshape(3, 3)
                kvec = np.array([K[0, 0], K[1, 1], K[0, 2], K[1, 2]])
                return kvec


def writePngMapFile(map_, filename):
    Image.fromarray(map_).save(filename)


def dispToBGR(disp, colormap=cv.COLORMAP_PLASMA):
    shape = disp.shape
    disp[np.isnan(disp)] = 0
    disp[disp < 0] = 0
    disp = disp.reshape(shape)
    disp = disp / disp.max()
    disp = (255 * disp).astype(np.uint8)
    disp = cv.applyColorMap(disp, colormap)
    return disp



import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from ptlflow.utils.utils import forward_interpolate_batch
from .extractor import BasicEncoder
from .attention import Attention1D
from .position import PositionEmbeddingSine
from .correlation import Correlation1D
from .update import BasicUpdateBlock
from .utils import coords_grid
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()

            flow_loss += i_weight * (valid[:, None] * i_loss).mean()

        return flow_loss


class Flow1D(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flow1d-chairs-75cd85a1.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flow1d-things-bcd92815.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flow1d-sintel-28a093d3.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flow1d-kitti-803a0181.ckpt",
        "highres": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flow1d-highres-7ab476dc.ckpt",
    }

    def __init__(
        self,
        downsample_factor: int = 8,
        feature_channels: int = 256,
        hidden_dim: int = 128,
        context_dim: int = 128,
        corr_radius: int = 32,
        iters: int = 32,
        mixed_precision: bool = False,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.downsample_factor = downsample_factor
        self.feature_channels = feature_channels
        self.hidden_dim = hidden_dim
        self.context_dim = context_dim
        self.corr_radius = corr_radius
        self.iters = iters
        self.mixed_precision = mixed_precision
        self.gamma = gamma
        self.max_flow = max_flow

        # feature network, context network, and update block
        self.fnet = BasicEncoder(
            output_dim=self.feature_channels,
            norm_fn="instance",
        )

        self.cnet = BasicEncoder(
            output_dim=self.hidden_dim + self.context_dim,
            norm_fn="batch",
        )

        # 1D attention
        corr_channels = (2 * self.corr_radius + 1) * 2

        self.attn_x = Attention1D(
            self.feature_channels,
            y_attention=False,
            double_cross_attn=True,
        )
        self.attn_y = Attention1D(
            self.feature_channels,
            y_attention=True,
            double_cross_attn=True,
        )

        # Update block
        self.update_block = BasicUpdateBlock(
            corr_channels=corr_channels,
            hidden_dim=self.hidden_dim,
            context_dim=self.context_dim,
            downsample_factor=self.downsample_factor,
        )

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow(self, img, downsample=None):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        n, c, h, w = img.shape
        downsample_factor = self.downsample_factor if downsample is None else downsample
        coords0 = coords_grid(
            n, h // downsample_factor, w // downsample_factor, dtype=img.dtype
        ).to(img.device)
        coords1 = coords_grid(
            n, h // downsample_factor, w // downsample_factor, dtype=img.dtype
        ).to(img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def learned_upflow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        n, _, h, w = flow.shape
        mask = mask.view(n, 1, 9, self.downsample_factor, self.downsample_factor, h, w)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(self.downsample_factor * flow, [3, 3], padding=1)
        up_flow = up_flow.view(n, 2, 9, 1, 1, h, w)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(
            n, 2, self.downsample_factor * h, self.downsample_factor * w
        )

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        # run the feature network
        feature1, feature2 = self.fnet([image1, image2])

        hdim = self.hidden_dim
        cdim = self.context_dim

        # position encoding
        pos_channels = self.feature_channels // 2
        pos_enc = PositionEmbeddingSine(pos_channels)

        position = pos_enc(feature1)  # [B, C, H, W]

        # 1D correlation
        feature2_x, attn_x = self.attn_x(feature1, feature2, position)
        corr_fn_y = Correlation1D(
            feature1,
            feature2_x,
            radius=self.corr_radius,
            x_correlation=False,
        )

        feature2_y, attn_y = self.attn_y(feature1, feature2, position)
        corr_fn_x = Correlation1D(
            feature1,
            feature2_y,
            radius=self.corr_radius,
            x_correlation=True,
        )

        # run the context network
        cnet = self.cnet(image1)  # list of feature pyramid, low scale to high scale

        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)

        coords0, coords1 = self.initialize_flow(image1)  # 1/8 resolution or 1/4

        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            forward_flow = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])
            coords1 = coords1 + forward_flow

        flow_predictions = []
        for itr in range(self.iters):
            coords1 = coords1.detach()  # stop gradient

            corr_x = corr_fn_x(coords1)
            corr_y = corr_fn_y(coords1)
            corr = torch.cat((corr_x, corr_y), dim=1)  # [B, 2(2R+1), H, W]

            flow = coords1 - coords0

            net, up_mask, delta_flow = self.update_block(
                net,
                inp,
                corr,
                flow,
                upsample=(self.training or itr == self.iters - 1),
            )

            coords1 = coords1 + delta_flow

            if self.training:
                # upsample predictions
                flow_up = self.learned_upflow(coords1 - coords0, up_mask)
                flow_up = self.postprocess_predictions(
                    flow_up, image_resizer, is_flow=True
                )
                flow_predictions.append(flow_up)
            elif itr == self.iters - 1:
                flow_up = self.learned_upflow(coords1 - coords0, up_mask)
                flow_up = self.postprocess_predictions(
                    flow_up, image_resizer, is_flow=True
                )

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": coords1 - coords0}

        return outputs


@register_model
@trainable
class flow1d(Flow1D):
    pass



""" DropBlock, DropPath

PyTorch implementations of DropBlock and DropPath (Stochastic Depth) regularization layers.

Papers:
DropBlock: A regularization method for convolutional networks (https://arxiv.org/abs/1810.12890)

Deep Networks with Stochastic Depth (https://arxiv.org/abs/1603.09382)

Code:
DropBlock impl inspired by two Tensorflow impl that I liked:
 - https://github.com/tensorflow/tpu/blob/master/models/official/resnet/resnet_model.py#L74
 - https://github.com/clovaai/assembled-cnn/blob/master/nets/blocks.py

Hacked together by / Copyright 2020 Ross Wightman
"""

import torch.nn as nn


def drop_path(
    x, drop_prob: float = 0.0, training: bool = False, scale_by_keep: bool = True
):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor.div_(keep_prob)
    return x * random_tensor


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training, self.scale_by_keep)

    def extra_repr(self):
        return f"drop_prob={round(self.drop_prob,3):0.3f}"



import torch
from .knn import knn_faiss_raw
from .utils import coords_grid, coords_grid_y_first


def normalize_coords(coords, H, W):
    """Normalize coordinates based on feature map shape. coords: [B, 2, N]"""
    one = coords.new_tensor(1)
    size = torch.stack([one * W, one * H])[None]
    center = size / 2
    scaling = size.max(1, keepdim=True).values * 0.5
    return (coords - center[:, :, None]) / scaling[:, :, None]


def compute_sparse_corr_init(fmap1, fmap2, k=32):
    """
    Compute a cost volume containing the k-largest hypotheses for each pixel.
    Output: corr_mink
    """
    B, C, H1, W1 = fmap1.shape
    H2, W2 = fmap2.shape[2:]
    N = H1 * W1

    fmap1, fmap2 = fmap1.view(B, C, -1), fmap2.view(B, C, -1)

    with torch.no_grad():
        _, indices = knn_faiss_raw(fmap1, fmap2, k)  # [B, k, H1*W1]

        indices_coord = indices.unsqueeze(1).expand(-1, 2, -1, -1)  # [B, 2, k, H1*W1]
        coords0 = (
            coords_grid_y_first(B, H2, W2)
            .view(B, 2, 1, -1)
            .expand(-1, -1, k, -1)
            .to(fmap1.device)
        )  # [B, 2, k, H1*W1]
        coords1 = coords0.gather(3, indices_coord)  # [B, 2, k, H1*W1]
        coords1 = coords1 - coords0

        # Append batch index
        batch_index = (
            torch.arange(B).view(B, 1, 1, 1).expand(-1, -1, k, N).type_as(coords1)
        )

    # Gather by indices from map2 and compute correlation volume
    fmap2 = fmap2.gather(2, indices.view(B, 1, -1).expand(-1, C, -1)).view(B, C, k, N)
    me_corr = torch.einsum("bcn,bckn->bkn", fmap1, fmap2).contiguous() / torch.sqrt(
        torch.tensor(C).float()
    )  # [B, k, H1*W1]

    return me_corr, coords0, coords1, batch_index  # coords: [B, 2, k, H1*W1]


if __name__ == "__main__":
    torch.manual_seed(0)

    for _ in range(100):
        fmap1 = torch.randn(8, 256, 92, 124).cuda()
        fmap2 = torch.randn(8, 256, 92, 124).cuda()
        corr_me = compute_sparse_corr_init(fmap1, fmap2, k=16)

    # corr_dense = corr(fmap1, fmap2)
    # corr_max = torch.max(corr_dense, dim=3)



from typing import Optional, Sequence

import torch
import torch.nn as nn

from ptlflow.utils.registry import register_model, trainable
from ptlflow.utils.utils import forward_interpolate_batch
from .optimizer import fetch_optimizer
from .MemFlowNet.corr import CorrBlock
from .memory_manager_skflow import MemoryManager
from .MemFlowNet.MemFlow import MemFlowNet
from .MemFlowNet.memory_util import *
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float, filter_epe: bool):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow
        self.filter_epe = filter_epe

    # def forward(self, flow_preds, flow_gt, valid):
    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        # print(flow_gt.shape, valid.shape, flow_preds[0].shape)
        # exit()

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]
        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=2).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)

            flow_pre = flow_preds[i]
            i_loss = (flow_pre - flow_gt).abs()

            _valid = valid[:, :, None]
            if self.filter_epe:
                loss_mag = torch.sum(i_loss**2, dim=2).sqrt()
                mask = loss_mag > 1000
                # print(mask.shape, _valid.shape)
                if torch.any(mask):
                    print(
                        "[Found extrem epe. Filtered out. Max is {}. Ratio is {}]".format(
                            torch.max(loss_mag), torch.mean(mask.float())
                        )
                    )
                    _valid = _valid & (~mask[:, :, None])

            flow_loss += i_weight * (_valid * i_loss).mean()

        return flow_loss


class MemFlow(BaseModel):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/memflow-things-90d0b74c.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/memflow-sintel-38621d84.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/memflow-kitti-ee6cbf09.ckpt",
        "spring": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/memflow-spring-7ee1b984.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        cnet: str = "basicencoder",
        fnet: str = "basicencoder",
        feat_dim: int = 256,
        gma: str = "GMA-SK2",
        corr_fn: str = "default",
        down_ratio: int = 8,
        decoder_depth: int = 15,
        cost_heads_num: int = 1,
        mem_every: int = 1,
        enable_long_term: bool = False,
        enable_long_term_count_usage: bool = True,
        max_mid_term_frames: int = 2,
        min_mid_term_frames: int = 2,
        num_prototypes: int = 128,
        max_long_term_elements: int = 10000,
        top_k: Optional[int] = None,
        critical_params: Sequence[str] = (
            "cnet",
            "fnet",
            "pretrain",
            "corr_levels",
            "decoder_depth",
            "train_avg_length",
        ),
        filter_epe: bool = False,
        pretrain: bool = True,
        train_avg_length: Optional[int] = None,
        gamma: float = 0.8,
        max_flow: float = 400,
        optimizer_name="adamw",
        scheduler_name="OneCycleLR",
        canonical_lr=2.1e-4,
        twins_lr_factor=2.1e-4,
        adam_decay=1e-4,
        adamw_decay=1e-4,
        epsilon=1e-8,
        num_steps=600000,
        anneal_strategy="linear",
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8, loss_fn=SequenceLoss(gamma, max_flow, filter_epe), **kwargs
        )

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.cnet = cnet
        self.fnet = fnet
        self.feat_dim = feat_dim
        self.gma = gma
        self.corr_fn = corr_fn
        self.down_ratio = down_ratio
        self.decoder_depth = decoder_depth
        self.cost_heads_num = cost_heads_num
        self.mem_every = mem_every
        self.enable_long_term = enable_long_term
        self.enable_long_term_count_usage = enable_long_term_count_usage
        self.max_mid_term_frames = max_mid_term_frames
        self.min_mid_term_frames = min_mid_term_frames
        self.num_prototypes = num_prototypes
        self.max_long_term_elements = max_long_term_elements
        self.top_k = top_k
        self.critical_params = critical_params
        self.filter_epe = filter_epe
        self.pretrain = pretrain
        self.train_avg_length = train_avg_length
        self.gamma = gamma
        self.max_flow = max_flow

        self.optimizer_name = (optimizer_name,)
        self.scheduler_name = (scheduler_name,)
        self.canonical_lr = (canonical_lr,)
        self.twins_lr_factor = (twins_lr_factor,)
        self.adam_decay = (adam_decay,)
        self.adamw_decay = (adamw_decay,)
        self.epsilon = (epsilon,)
        self.num_steps = (num_steps,)
        self.anneal_strategy = (anneal_strategy,)

        self.network = MemFlowNet(
            cnet=cnet,
            fnet=fnet,
            corr_levels=corr_levels,
            corr_radius=corr_radius,
            decoder_depth=decoder_depth,
            pretrain=pretrain,
            gma=gma,
            corr_fn=corr_fn,
            train_avg_length=train_avg_length,
        )

        if self.train_avg_length is None:
            train_avg_length = 6750
            print(
                f"WARNING: --train_avg_length is not provided and it cannot be loaded from the checkpoint either. It will be set as {train_avg_length}, but this may not be the optimal value."
            )
            self.train_avg_length = train_avg_length

        self.clear_memory()

    def clear_memory(self):
        self.curr_ti = -1
        self.last_mem_ti = -self.mem_every
        self.memory = MemoryManager(
            train_avg_length=self.train_avg_length,
            enable_long_term=self.enable_long_term,
            enable_long_term_count_usage=self.enable_long_term_count_usage,
            top_k=self.top_k,
            max_mid_term_frames=self.max_mid_term_frames,
            min_mid_term_frames=self.min_mid_term_frames,
        )

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        if (
            "meta" in inputs
            and "is_seq_start" in inputs["meta"]
            and inputs["meta"]["is_seq_start"]
        ):
            self.clear_memory()

        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        flow_init = None
        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            flow_init = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])

        # image: 1*2*3*H*W
        self.curr_ti += 1

        end = True
        if "meta" in inputs and "is_seq_end" in inputs["meta"]:
            end = inputs["meta"]["is_seq_end"]
        is_mem_frame = (self.curr_ti - self.last_mem_ti >= self.mem_every) and (not end)

        # B, C, H, W
        query, key, net, inp = self.network.encode_context(images[:, 0, ...])
        # B, T, C, H, W
        coords0, coords1, fmaps = self.network.encode_features(
            images, flow_init=flow_init
        )

        # predict flow
        corr_fn = CorrBlock(fmaps[:, 0, ...], fmaps[:, 1, ...], num_levels=4, radius=4)

        for itr in range(self.decoder_depth):
            coords1 = coords1.detach()
            corr = corr_fn(coords1)  # index correlation volume
            flow = coords1 - coords0
            (
                motion_features,
                current_value,
            ) = self.network.update_block.get_motion_and_value(flow, corr)
            # get global motion
            memory_readout = self.memory.match_memory(
                query, key, current_value, scale=self.network.att.scale
            )
            motion_features_global = (
                motion_features
                + self.network.update_block.aggregator.gamma * memory_readout
            )
            net, up_mask, delta_flow = self.network.update_block(
                net, inp, motion_features, motion_features_global
            )
            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow
        # upsample predictions
        flow_up = self.network.upsample_flow(coords1 - coords0, up_mask)
        flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)

        # save as memory if needed
        if is_mem_frame:
            self.memory.add_memory(key, current_value)
            self.last_mem_ti = self.curr_ti

        # if self.training:
        #     outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        # else:
        outputs = {"flows": flow_up[:, None], "flow_small": coords1 - coords0}

        return outputs

    def configure_optimizers(self):
        optimizer, lr_scheduler = fetch_optimizer(
            model=self,
            optimizer_name=self.optimizer_name,
            scheduler_name=self.scheduler_name,
            canonical_lr=self.canonical_lr,
            twins_lr_factor=self.twins_lr_factor,
            adam_decay=self.adam_decay,
            adamw_decay=self.adamw_decay,
            epsilon=self.epsilon,
            num_steps=self.num_steps,
            anneal_strategy=self.anneal_strategy,
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": {"scheduler": lr_scheduler, "interval": "step"},
        }


class MemFlowT(MemFlow):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/memflow_t-things-6028d89f.ckpt",
        "things_kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/memflow_t-things_kitti-542e0a1c.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/memflow_t-sintel-d2df0424.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/memflow_t-kitti-9eeabb65.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        cnet: str = "twins",
        fnet: str = "twins",
        feat_dim: int = 256,
        gma: str = "GMA-SK2",
        corr_fn: str = "default",
        down_ratio: int = 8,
        decoder_depth: int = 15,
        cost_heads_num: int = 1,
        mem_every: int = 1,
        enable_long_term: bool = False,
        enable_long_term_count_usage: bool = True,
        max_mid_term_frames: int = 2,
        min_mid_term_frames: int = 2,
        num_prototypes: int = 128,
        max_long_term_elements: int = 10000,
        top_k: int | None = None,
        critical_params: Sequence[str] = (
            "cnet",
            "fnet",
            "pretrain",
            "corr_levels",
            "decoder_depth",
            "train_avg_length",
        ),
        filter_epe: bool = False,
        pretrain: bool = True,
        train_avg_length: int | None = None,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        super().__init__(
            corr_levels,
            corr_radius,
            cnet,
            fnet,
            feat_dim,
            gma,
            corr_fn,
            down_ratio,
            decoder_depth,
            cost_heads_num,
            mem_every,
            enable_long_term,
            enable_long_term_count_usage,
            max_mid_term_frames,
            min_mid_term_frames,
            num_prototypes,
            max_long_term_elements,
            top_k,
            critical_params,
            filter_epe,
            pretrain,
            train_avg_length,
            gamma,
            max_flow,
            **kwargs,
        )


@register_model
@trainable
class memflow(MemFlow):
    pass


@register_model
@trainable
class memflow_t(MemFlowT):
    pass



"""
Portions of this code copyright 2017, Clement Pinard
"""

from argparse import Namespace

import torch
import torch.nn as nn
from torch.nn import init

from .submodules import *
from .flownet_base import FlowNetBase


class FlowNetFusion(FlowNetBase):
    def __init__(
        self,
        div_flow: float = 20.0,
        input_channels: int = 6,
        batch_norm: bool = False,
        loss_start_scale: int = 1,
        loss_num_scales: int = 3,
        loss_base_weight: float = 0.32,
        loss_norm: str = "L2",
        **kwargs,
    ):
        super(FlowNetFusion, self).__init__(
            div_flow=div_flow,
            input_channels=input_channels,
            batch_norm=batch_norm,
            loss_start_scale=loss_start_scale,
            loss_num_scales=loss_num_scales,
            loss_base_weight=loss_base_weight,
            loss_norm=loss_norm,
            **kwargs,
        )

        self.conv0 = conv(self.batch_norm, 11, 64)
        self.conv1 = conv(self.batch_norm, 64, 64, stride=2)
        self.conv1_1 = conv(self.batch_norm, 64, 128)
        self.conv2 = conv(self.batch_norm, 128, 128, stride=2)
        self.conv2_1 = conv(self.batch_norm, 128, 128)

        self.deconv1 = deconv(128, 32)
        self.deconv0 = deconv(162, 16)

        self.inter_conv1 = i_conv(self.batch_norm, 162, 32)
        self.inter_conv0 = i_conv(self.batch_norm, 82, 16)

        self.predict_flow2 = predict_flow(128)
        self.predict_flow1 = predict_flow(32)
        self.predict_flow0 = predict_flow(16)

        self.upsampled_flow2_to_1 = nn.ConvTranspose2d(2, 2, 4, 2, 1)
        self.upsampled_flow1_to_0 = nn.ConvTranspose2d(2, 2, 4, 2, 1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

    def forward(self, inputs):
        x = inputs["images"]

        x = x.view(x.shape[0], x.shape[1] * x.shape[2], x.shape[3], x.shape[4])
        out_conv0 = self.conv0(x)
        out_conv1 = self.conv1_1(self.conv1(out_conv0))
        out_conv2 = self.conv2_1(self.conv2(out_conv1))

        flow2 = self.predict_flow2(out_conv2)
        flow2_up = self.upsampled_flow2_to_1(flow2)
        out_deconv1 = self.deconv1(out_conv2)

        concat1 = torch.cat((out_conv1, out_deconv1, flow2_up), 1)
        out_interconv1 = self.inter_conv1(concat1)
        flow1 = self.predict_flow1(out_interconv1)
        flow1_up = self.upsampled_flow1_to_0(flow1)
        out_deconv0 = self.deconv0(concat1)

        concat0 = torch.cat((out_conv0, out_deconv0, flow1_up), 1)
        out_interconv0 = self.inter_conv0(concat0)
        flow0 = self.predict_flow0(out_interconv0)

        outputs = {}

        if self.training:
            outputs["flow_preds"] = [flow0, flow1, flow2]
            outputs["flows"] = flow0[:, None]
        else:
            outputs["flows"] = flow0[:, None]

        return outputs



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import einsum

from einops import rearrange

from ...utils import coords_grid
from .attention import (
    MultiHeadAttention,
    LinearPositionEmbeddingSine,
    ExpPositionEmbeddingSine,
)
from ..encoders import twins_svt_large, convnext_large
from .cnn import BasicEncoder

from ...timm0412.models.layers import DropPath


class PatchEmbed(nn.Module):
    def __init__(
        self, patch_embed, use_rpe, patch_size=16, in_chans=1, embed_dim=64, pe="linear"
    ):
        super().__init__()
        self.use_rpe = use_rpe
        self.patch_size = patch_size
        self.dim = embed_dim
        self.pe = pe

        if patch_embed == "no_relu":
            self.proj = nn.Sequential(
                nn.Conv2d(in_chans, embed_dim // 4, kernel_size=6, stride=2, padding=2),
                nn.Conv2d(
                    embed_dim // 4, embed_dim // 2, kernel_size=6, stride=2, padding=2
                ),
                nn.Conv2d(
                    embed_dim // 2, embed_dim, kernel_size=6, stride=2, padding=2
                ),
            )
        elif patch_embed == "single":
            # assert patch_size == 8
            if patch_size == 8:
                self.proj = nn.ModuleList(
                    [
                        nn.Conv2d(
                            in_chans, embed_dim // 4, kernel_size=6, stride=2, padding=2
                        ),
                        nn.ReLU(),
                        nn.Conv2d(
                            embed_dim // 4,
                            embed_dim // 2,
                            kernel_size=6,
                            stride=2,
                            padding=2,
                        ),
                        nn.ReLU(),
                        nn.Conv2d(
                            embed_dim // 2,
                            embed_dim,
                            kernel_size=6,
                            stride=2,
                            padding=2,
                        ),
                    ]
                )
            elif patch_size == 4:
                self.proj = nn.Sequential(
                    nn.Conv2d(
                        in_chans, embed_dim // 4, kernel_size=6, stride=2, padding=2
                    ),
                    nn.ReLU(),
                    nn.Conv2d(
                        embed_dim // 4, embed_dim, kernel_size=6, stride=2, padding=2
                    ),
                )
            else:
                print(f"patch size = {patch_size} is unacceptable.")

        self.ffn_with_coord = nn.Sequential(
            nn.Conv2d(embed_dim + 64, embed_dim + 64, kernel_size=1),
            nn.ReLU(),
            nn.Conv2d(embed_dim + 64, embed_dim + 64, kernel_size=1),
        )
        self.norm = nn.LayerNorm(embed_dim + 64)

    def forward(
        self, x, mask_for_patch1=None, mask_for_patch2=None, mask_for_patch3=None
    ):
        B, C, H, W = x.shape  # C == 1

        pad_l = pad_t = 0
        pad_r = (self.patch_size - W % self.patch_size) % self.patch_size
        pad_b = (self.patch_size - H % self.patch_size) % self.patch_size
        x = F.pad(x, (pad_l, pad_r, pad_t, pad_b))

        masks = [mask_for_patch1, mask_for_patch2, mask_for_patch3]
        for idx, layer in enumerate(self.proj):
            if idx % 2 == 0:
                if masks[idx // 2] is not None:
                    x = x * (1 - masks[idx // 2])
            x = layer(x)

        out_size = x.shape[2:]

        patch_coord = (
            coords_grid(B, out_size[0], out_size[1], dtype=x.dtype, device=x.device)
            * self.patch_size
            + self.patch_size / 2
        )  # in feature coordinate space
        if self.use_rpe:
            center_coord = coords_grid(1, H, W, dtype=x.dtype, device=x.device)
            center_coord = (
                center_coord.permute(2, 3, 1, 0)
                .reshape(H * W, 2, 1, 1)
                .repeat(B // (H * W), 1, 1, 1)
            )
            patch_coord = patch_coord - center_coord

        patch_coord = patch_coord.view(B, 2, -1).permute(0, 2, 1)
        if self.pe == "linear":
            patch_coord_enc = LinearPositionEmbeddingSine(patch_coord, dim=64)
        elif self.pe == "exp":
            patch_coord_enc = ExpPositionEmbeddingSine(patch_coord, dim=64)
        patch_coord_enc = patch_coord_enc.permute(0, 2, 1).view(
            B, -1, out_size[0], out_size[1]
        )

        x_pe = torch.cat([x, patch_coord_enc], dim=1)
        x = self.ffn_with_coord(x_pe)

        x = self.norm(x.flatten(2).transpose(1, 2))

        return x, out_size


from .twins import Block


class VerticalSelfAttentionLayer(nn.Module):
    def __init__(
        self,
        dim,
        droppath,
        vert_c_dim,
        encoder_latent_dim,
        num_heads=8,
        dropout=0.0,
    ):
        super(VerticalSelfAttentionLayer, self).__init__()
        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        embed_dim = dim
        mlp_ratio = 4
        ws = 7
        sr_ratio = 4
        dpr = droppath
        drop_rate = dropout
        attn_drop_rate = 0.0

        self.local_block = Block(
            dim=embed_dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            drop=drop_rate,
            attn_drop=attn_drop_rate,
            drop_path=dpr,
            sr_ratio=sr_ratio,
            ws=ws,
            with_rpe=True,
            vert_c_dim=vert_c_dim,
            encoder_latent_dim=encoder_latent_dim,
        )
        self.global_block = Block(
            dim=embed_dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
            drop=drop_rate,
            attn_drop=attn_drop_rate,
            drop_path=dpr,
            sr_ratio=sr_ratio,
            ws=1,
            with_rpe=True,
            vert_c_dim=vert_c_dim,
            encoder_latent_dim=encoder_latent_dim,
        )

    def forward(self, x, size, context=None):
        x = self.local_block(x, size, context)
        x = self.global_block(x, size, context)

        return x


class SelfAttentionLayer(nn.Module):
    def __init__(
        self,
        dim,
        droppath,
        num_heads=8,
        proj_drop=0.0,
        drop_path=0.0,
        dropout=0.0,
    ):
        super(SelfAttentionLayer, self).__init__()
        assert (
            dim % num_heads == 0
        ), f"dim {dim} should be divided by num_heads {num_heads}."

        proj_drop = drop_path = droppath

        self.dim = dim
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.norm1 = nn.LayerNorm(dim)
        self.norm2 = nn.LayerNorm(dim)
        self.multi_head_attn = MultiHeadAttention(dim, num_heads)
        self.q, self.k, self.v = (
            nn.Linear(dim, dim, bias=True),
            nn.Linear(dim, dim, bias=True),
            nn.Linear(dim, dim, bias=True),
        )

        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(dim, dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(dim, dim),
            nn.Dropout(dropout),
        )

    def forward(self, x):
        """
        x: [BH1W1, H3W3, D]
        """
        short_cut = x
        x = self.norm1(x)

        q, k, v = self.q(x), self.k(x), self.v(x)

        x = self.multi_head_attn(q, k, v)

        x = self.proj(x)
        x = short_cut + self.proj_drop(x)

        x = x + self.drop_path(self.ffn(self.norm2(x)))

        return x


class CostPerceiverEncoder(nn.Module):
    def __init__(
        self,
        cost_heads_num,
        vertical_encoder_attn,
        cost_latent_token_num,
        cost_latent_dim,
        cost_encoder_res,
        mask_ratio,
        patch_size,
        cost_latent_input_dim,
        pe,
        encoder_depth,
        cross_attn,
        dropout,
        vert_c_dim,
        patch_embed,
        use_rpe,
        droppath,
        encoder_latent_dim,
        attn_dim,
        expand_factor,
    ):
        super(CostPerceiverEncoder, self).__init__()
        self.cost_heads_num = cost_heads_num
        self.vertical_encoder_attn = vertical_encoder_attn
        self.cost_latent_token_num = cost_latent_token_num
        self.cost_encoder_res = cost_encoder_res
        self.mask_ratio = mask_ratio

        self.patch_size = patch_size
        self.patch_embed = PatchEmbed(
            patch_embed=patch_embed,
            use_rpe=use_rpe,
            in_chans=cost_heads_num,
            patch_size=self.patch_size,
            embed_dim=cost_latent_input_dim,
            pe=pe,
        )

        self.depth = encoder_depth

        self.latent_tokens = nn.Parameter(
            torch.randn(1, cost_latent_token_num, cost_latent_dim)
        )

        query_token_dim, tgt_token_dim = (
            cost_latent_dim,
            cost_latent_input_dim * 2,
        )
        qk_dim, v_dim = query_token_dim, query_token_dim

        if cross_attn == "all":
            from .crossattentionlayer import CrossAttentionLayer

            self.input_layer = CrossAttentionLayer(
                qk_dim, v_dim, query_token_dim, tgt_token_dim, dropout=dropout
            )
        elif cross_attn == "part":
            from .crossattentionlayer import CrossAttentionLayer_two_level

            self.input_layer = CrossAttentionLayer_two_level(
                qk_dim, v_dim, query_token_dim, tgt_token_dim, dropout=dropout
            )
        elif cross_attn == "rep":
            from .crossattentionlayer import CrossAttentionLayer_two_level_rep

            self.input_layer = CrossAttentionLayer_two_level_rep(
                qk_dim, v_dim, query_token_dim, tgt_token_dim, dropout=dropout
            )
        elif cross_attn == "k3s2":
            from .crossattentionlayer import CrossAttentionLayer_convk3s2

            self.input_layer = CrossAttentionLayer_convk3s2(
                qk_dim, v_dim, query_token_dim, tgt_token_dim, dropout=dropout
            )
        elif cross_attn == "34":
            print("[Using 34 crossattention layer]")
            from .crossattentionlayer import CrossAttentionLayer_34

            self.input_layer = CrossAttentionLayer_34(
                qk_dim, v_dim, query_token_dim, tgt_token_dim, dropout=dropout
            )

        self.encoder_layers = nn.ModuleList(
            [
                SelfAttentionLayer(cost_latent_dim, droppath=droppath, dropout=dropout)
                for idx in range(self.depth)
            ]
        )

        if vertical_encoder_attn == "twins":
            self.vertical_encoder_layers = nn.ModuleList(
                [
                    VerticalSelfAttentionLayer(
                        cost_latent_dim,
                        droppath=droppath,
                        vert_c_dim=vert_c_dim,
                        encoder_latent_dim=encoder_latent_dim,
                        dropout=dropout,
                    )
                    for idx in range(self.depth)
                ]
            )
        elif vertical_encoder_attn == "NA":
            from .NA import selfattentionlayer_nat

            self.vertical_encoder_layers = nn.ModuleList(
                [
                    selfattentionlayer_nat(
                        dropout=dropout,
                        droppath=droppath,
                        attn_dim=attn_dim,
                        expand_factor=expand_factor,
                        encoder_latent_dim=encoder_latent_dim,
                        vert_c_dim=vert_c_dim,
                        cost_latent_dim=cost_latent_dim,
                    )
                    for idx in range(self.depth)
                ]
            )
        elif vertical_encoder_attn == "NA-twins":
            from .NA import NATwins

            assert vert_c_dim > 0, "Only support vert_c_dim>0 for NA-Twins"
            print("[Using NA-twins vertical attention layers]")
            self.vertical_encoder_layers = nn.ModuleList(
                [
                    NATwins(
                        cost_latent_dim=cost_latent_dim,
                        dropout=dropout,
                        droppath=droppath,
                        vert_c_dim=vert_c_dim,
                        encoder_latent_dim=encoder_latent_dim,
                        attn_dim=attn_dim,
                        expand_factor=expand_factor,
                    )
                    for idx in range(self.depth)
                ]
            )

    def random_masking(self, x, mask_ratio, mask=None):
        B, _, H, W = x.shape

        pad_l = pad_t = 0
        pad_r = (self.patch_size - W % self.patch_size) % self.patch_size
        pad_b = (self.patch_size - H % self.patch_size) % self.patch_size
        H = H + pad_b
        W = W + pad_r

        # number of keys for crossattentionlayer
        H_down = H // 8
        W_down = W // 8

        L = H_down * W_down
        len_keep = int(L * (1 - mask_ratio))

        if mask is not None:
            noise = mask.reshape(B, L)
        else:
            # print("random mask")
            noise = torch.rand(B, L, device=x.device)  # noise in [0, 1]

        # sort noise for each sample
        ids_shuffle = torch.argsort(
            noise, dim=1
        )  # ascend: small is keep, large is remove
        ids_restore = torch.argsort(ids_shuffle, dim=1)

        # keep the first subset
        ids_keep = ids_shuffle[:, :len_keep]
        # generate the binary mask: 0 is keep, 1 is remove
        mask = torch.ones([B, L], device=x.device)
        mask[:, :len_keep] = 0
        # unshuffle to get the binary mask
        mask_for_keys = torch.gather(mask, dim=1, index=ids_restore)

        mask_for_patch1 = (
            mask_for_keys.reshape(-1, H_down, W_down)
            .unsqueeze(-1)
            .repeat(1, 1, 1, 64)
            .reshape(-1, H_down, W_down, 8, 8)
            .permute(0, 1, 3, 2, 4)
            .reshape(B, H_down * 8, W_down * 8)
            .unsqueeze(1)
        )
        mask_for_patch2 = (
            mask_for_keys.reshape(-1, H_down, W_down)
            .unsqueeze(-1)
            .repeat(1, 1, 1, 16)
            .reshape(-1, H_down, W_down, 4, 4)
            .permute(0, 1, 3, 2, 4)
            .reshape(B, H_down * 4, W_down * 4)
            .unsqueeze(1)
        )
        mask_for_patch3 = (
            mask_for_keys.reshape(-1, H_down, W_down)
            .unsqueeze(-1)
            .repeat(1, 1, 1, 4)
            .reshape(-1, H_down, W_down, 2, 2)
            .permute(0, 1, 3, 2, 4)
            .reshape(B, H_down * 2, W_down * 2)
            .unsqueeze(1)
        )

        return (
            ids_keep,
            mask_for_keys,
            mask_for_patch1,
            mask_for_patch2,
            mask_for_patch3,
            ids_restore,
        )

    def forward(self, cost_volume, data, context=None):
        B, heads, H1, W1, H2, W2 = cost_volume.shape
        cost_maps = (
            cost_volume.permute(0, 2, 3, 1, 4, 5)
            .contiguous()
            .view(B * H1 * W1, self.cost_heads_num, H2, W2)
        )
        data["cost_maps"] = cost_maps

        x, size = self.patch_embed(cost_maps)  # B*H1*W1, size[0]*size[1], C
        data["H3W3"] = size
        H3, W3 = size

        cost_patches = x

        x = self.input_layer(self.latent_tokens, x, size)

        short_cut = x

        for idx, layer in enumerate(self.encoder_layers):
            x = layer(x)
            if self.vertical_encoder_attn is not None:
                x = (
                    x.view(B, H1 * W1, self.cost_latent_token_num, -1)
                    .permute(0, 2, 1, 3)
                    .reshape(B * self.cost_latent_token_num, H1 * W1, -1)
                )
                x = self.vertical_encoder_layers[idx](x, (H1, W1), context)
                x = (
                    x.view(B, self.cost_latent_token_num, H1 * W1, -1)
                    .permute(0, 2, 1, 3)
                    .reshape(B * H1 * W1, self.cost_latent_token_num, -1)
                )

        if self.cost_encoder_res is True:
            x = x + short_cut

        _B, _HW, _C = cost_patches.shape
        cost_patches = cost_patches.reshape(_B, H3, W3, _C).permute(0, 3, 1, 2)

        return x, cost_patches

    def pretrain_forward(
        self, cost_volume_outter, cost_volume, data, context=None, mask=None
    ):
        B, heads, H1, W1, H2, W2 = cost_volume_outter.shape
        cost_maps = (
            cost_volume_outter.permute(0, 2, 3, 1, 4, 5)
            .contiguous()
            .view(B * H1 * W1, self.cost_heads_num, H2, W2)
        )
        data["cost_maps_outter"] = cost_maps

        B, heads, H1, W1, H2, W2 = cost_volume.shape
        cost_maps = (
            cost_volume.permute(0, 2, 3, 1, 4, 5)
            .contiguous()
            .view(B * H1 * W1, self.cost_heads_num, H2, W2)
        )
        data["cost_maps"] = cost_maps

        (
            ids_keep,
            mask_for_keys,
            mask_for_patch1,
            mask_for_patch2,
            mask_for_patch3,
            ids_restore,
        ) = self.random_masking(cost_maps, self.mask_ratio, mask)
        # ids_keep = mask_for_keys = mask_for_patch1 = mask_for_patch2 = mask_for_patch3 = ids_restore = None

        x, size = self.patch_embed(
            cost_maps, mask_for_patch1, mask_for_patch2, mask_for_patch3
        )  # B*H1*W1, size[0]*size[1], C

        data["H3W3"] = size
        H3, W3 = size

        cost_patches = x

        x = self.input_layer(self.latent_tokens, x, size, ids_keep)

        short_cut = x

        for idx, layer in enumerate(self.encoder_layers):
            x = layer(x)
            if self.vertical_encoder_attn is not None:
                x = (
                    x.view(B, H1 * W1, self.cost_latent_token_num, -1)
                    .permute(0, 2, 1, 3)
                    .reshape(B * self.cost_latent_token_num, H1 * W1, -1)
                )
                x = self.vertical_encoder_layers[idx](x, (H1, W1), context)
                x = (
                    x.view(B, self.cost_latent_token_num, H1 * W1, -1)
                    .permute(0, 2, 1, 3)
                    .reshape(B * H1 * W1, self.cost_latent_token_num, -1)
                )

        if self.cost_encoder_res is True:
            x = x + short_cut

        _B, _HW, _C = cost_patches.shape
        cost_patches = cost_patches.reshape(_B, H3, W3, _C).permute(0, 3, 1, 2)

        return x, cost_patches


class MemoryEncoder(nn.Module):
    def __init__(
        self,
        cost_heads_num,
        use_convertor,
        r_16,
        crop_cost_volume,
        del_layers,
        H_offset,
        W_offset,
        fnet,
        pretrain,
        pretrain_mode,
        encoder_latent_dim,
        vertical_encoder_attn,
        cost_latent_token_num,
        cost_latent_dim,
        cost_encoder_res,
        mask_ratio,
        patch_size,
        cost_latent_input_dim,
        pe,
        encoder_depth,
        cross_attn,
        dropout,
        vert_c_dim,
        patch_embed,
        use_rpe,
        droppath,
        attn_dim,
        expand_factor,
    ):
        super(MemoryEncoder, self).__init__()

        self.cost_heads_num = cost_heads_num
        self.use_convertor = use_convertor
        self.r_16 = r_16
        self.crop_cost_volume = crop_cost_volume
        self.H_offset = H_offset
        self.W_offset = W_offset

        if fnet == "twins":
            self.feat_encoder = twins_svt_large(
                pretrained=pretrain, del_layers=del_layers
            )
        elif fnet == "basicencoder":
            self.feat_encoder = BasicEncoder(output_dim=256, norm_fn="instance")
        elif fnet == "convnext":
            self.feat_encoder = convnext_large(pretrained=pretrain)
        else:
            exit()

        if pretrain_mode:
            print("[In pretrain mode, freeze feature encoder]")
            for param in self.feat_encoder.parameters():
                param.requires_grad = False

        if use_convertor:
            self.channel_convertor = nn.Conv2d(
                encoder_latent_dim, 256, 1, padding=0, bias=False
            )

        self.cost_perceiver_encoder = CostPerceiverEncoder(
            cost_heads_num=cost_heads_num,
            vertical_encoder_attn=vertical_encoder_attn,
            cost_latent_token_num=cost_latent_token_num,
            cost_latent_dim=cost_latent_dim,
            cost_encoder_res=cost_encoder_res,
            mask_ratio=mask_ratio,
            patch_size=patch_size,
            cost_latent_input_dim=cost_latent_input_dim,
            pe=pe,
            encoder_depth=encoder_depth,
            cross_attn=cross_attn,
            dropout=dropout,
            vert_c_dim=vert_c_dim,
            patch_embed=patch_embed,
            use_rpe=use_rpe,
            droppath=droppath,
            encoder_latent_dim=encoder_latent_dim,
            attn_dim=attn_dim,
            expand_factor=expand_factor,
        )

        if pretrain_mode and crop_cost_volume:
            print(
                "[H_offset is {}, W_offset is {}, and crop_cost_volume to get inner cost volume]".format(
                    H_offset, W_offset
                )
            )

    def corr(self, fmap1, fmap2):
        batch, dim, ht, wd = fmap1.shape
        _, _, ht2, wd2 = fmap2.shape

        fmap1 = rearrange(
            fmap1, "b (heads d) h w -> b heads (h w) d", heads=self.cost_heads_num
        )
        fmap2 = rearrange(
            fmap2, "b (heads d) h w -> b heads (h w) d", heads=self.cost_heads_num
        )
        corr = einsum("bhid, bhjd -> bhij", fmap1, fmap2)
        corr = corr.view(batch, self.cost_heads_num, ht, wd, ht2, wd2)

        return corr

    def corr_16(self, fmap1, fmap2):
        batch, dim, ht, wd = fmap1.shape
        _, _, ht2, wd2 = fmap2.shape

        # fmap1 = F.avg_pool2d(fmap1, kernel_size=2, stride=2, padding=0)
        fmap1 = fmap1[:, :, ::2, ::2]

        fmap1 = rearrange(
            fmap1, "b (heads d) h w -> b heads (h w) d", heads=self.cost_heads_num
        )
        fmap2 = rearrange(
            fmap2, "b (heads d) h w -> b heads (h w) d", heads=self.cost_heads_num
        )
        corr = einsum("bhid, bhjd -> bhij", fmap1, fmap2)
        corr = corr.view(batch, self.cost_heads_num, ht // 2, wd // 2, ht2, wd2)

        return corr

    def forward(self, img1, img2, data, context=None):
        feat_s, _ = self.feat_encoder(img1)
        feat_t, _ = self.feat_encoder(img2)

        feat_s_16 = None
        feat_t_16 = None

        B, C, H, W = feat_s.shape
        size = (H, W)

        if self.use_convertor:
            feat_s = self.channel_convertor(feat_s)
            feat_t = self.channel_convertor(feat_t)

        cost_volume = self.corr(feat_s, feat_t)
        if self.r_16 > 0:
            cost_volume_16 = self.corr_16(feat_s_16, feat_t_16)
            B, heads, H1, W1, H2, W2 = cost_volume_16.shape
            cost_maps = (
                cost_volume_16.permute(0, 2, 3, 1, 4, 5)
                .contiguous()
                .view(B * H1 * W1, self.cost_heads_num, H2, W2)
            )
            data["cost_maps_16"] = cost_maps

        x, cost_patches = self.cost_perceiver_encoder(cost_volume, data, context)

        return x, cost_patches, feat_s_16, feat_t_16

    def pretrain_forward(
        self, img1, img2, img1_inner, img2_inner, data, context=None, mask=None
    ):
        feat_t, _ = self.feat_encoder(img2)

        feat_s_inner, _ = self.feat_encoder(img1_inner)
        feat_t_inner, _ = self.feat_encoder(img2_inner)

        cost_volume = self.corr(feat_s_inner, feat_t)
        if self.crop_cost_volume:
            H_border = self.H_offset // 8
            W_border = self.W_offset // 8
            cost_volume_inner = cost_volume[
                :, :, :, :, H_border:-H_border, W_border:-W_border
            ]
        else:
            cost_volume_inner = self.corr(feat_s_inner, feat_t_inner)
        x, cost_patches = self.cost_perceiver_encoder.pretrain_forward(
            cost_volume, cost_volume_inner, data, context, mask=mask
        )

        return x, cost_patches



# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from pathlib import Path
import shutil

import cv2 as cv
from jsonargparse import ArgumentParser
import numpy as np

import infer
import ptlflow

TEST_MODEL = "raft_small"


def test_infer(tmp_path: Path) -> None:
    _create_images(tmp_path)

    model_ref = ptlflow.get_model_reference(TEST_MODEL)

    parser = ArgumentParser(parents=[infer._init_parser()])
    parser.add_class_arguments(model_ref, "model")
    args = parser.parse_args([])

    args.input_path = [str(tmp_path / "img1.png"), str(tmp_path / "img2.png")]
    args.write_outputs = True
    args.output_path = tmp_path

    args.flow_format = "flo"
    model = ptlflow.get_model(TEST_MODEL, None, args)
    infer.infer(args, model)
    assert (tmp_path / "flows/test_infer0/img1.flo").exists()
    assert (tmp_path / "flows_viz/test_infer0/img1.png").exists()

    args.flow_format = "png"
    model = ptlflow.get_model(TEST_MODEL, None, args)
    infer.infer(args, model)
    assert (tmp_path / "flows/test_infer0/img1.png").exists()
    assert (tmp_path / "flows_viz/test_infer0/img1.png").exists()

    shutil.rmtree(tmp_path)


def _create_images(tmp_path: Path) -> None:
    for i in range(2):
        img = np.random.randint(0, 255, (400, 400, 3), np.uint8)
        cv.imwrite(str(tmp_path / f"img{i+1}.png"), img)



import torch
import torch.nn.functional as F


MAX_FLOW = 400


def coords_grid(batch, ht, wd, dtype, device):
    coords = torch.meshgrid(
        torch.arange(ht, dtype=dtype, device=device),
        torch.arange(wd, dtype=dtype, device=device),
        indexing="ij",
    )
    coords = torch.stack(coords[::-1], dim=0)
    return coords[None].repeat(batch, 1, 1, 1)


def flow_to_warp(flow):
    """
    Compute the warp from the flow field
    Args:
        flow: optical flow shape [B, 2, H, W]
    Returns:
        warp: the endpoints of the estimated flow. shape [B, H, W, 2]
    """
    flow = flow.permute(0, 2, 3, 1)
    B, H, W, _ = flow.size()
    grid = coords_grid(B, H, W, dtype=flow.dtype, device=flow.device)
    if flow.is_cuda:
        grid = grid.cuda()
    grid = grid.permute(0, 2, 3, 1)
    warp = grid + flow
    return warp


def warp(x, flo, mode="bilinear"):
    H, W = flo.shape[-2:]
    vgrid = flow_to_warp(flo)
    vgrid[:, :, :, 0] = 2.0 * vgrid[:, :, :, 0] / max(W - 1, 1) - 1.0
    vgrid[:, :, :, 1] = 2.0 * vgrid[:, :, :, 1] / max(H - 1, 1) - 1.0
    if mode == "bilinear":
        output = F.grid_sample(x, vgrid, mode=mode, align_corners=True)
    else:
        output = F.grid_sample(x, vgrid, mode=mode)
    return output


def mask_invalid(coords, pad_h=0, pad_w=0):
    """
    Mask coordinates outside of the image.

    Valid = 1, invalid = 0.

    Args:
        coords: a 4D float tensor of image coordinates. [B, H, W, 2]
        pad_h: int, the amount of padding applied to the top of the image
        pad_w: int, the amount of padding applied to the left of the image

    Returns:
        The mask showing which coordinates are valid. [B, 1, H, W]
    """
    pad_h = float(pad_h)
    pad_w = float(pad_w)
    coords_rank = len(coords.shape)
    if coords_rank != 4:
        raise NotImplementedError()
    max_height = float(coords.shape[-3] - 1)
    max_width = float(coords.shape[-2] - 1)
    mask = torch.logical_and(
        torch.logical_and(coords[:, :, :, 0] >= pad_w, coords[:, :, :, 0] <= max_width),
        torch.logical_and(
            coords[:, :, :, 1] >= pad_h, coords[:, :, :, 1] <= max_height
        ),
    )
    mask = mask.to(dtype=coords.dtype)[:, None, :, :]
    return mask


def compute_range_map(flow):
    """Using backward flow to compute the range map"""
    # permute flow from [B, 2, H, W] to shape [B, H, W, 2]
    batch_size, _, input_height, input_width = flow.size()

    coords = flow_to_warp(flow)

    # split coordinates into an integer part and a float offset for interpolation.
    coords_floor = torch.floor(coords)
    coords_offset = coords - coords_floor
    coords_floor = coords_floor.to(torch.int32)

    # Define a batch offset for flattened indexes into all pixels
    batch_range = torch.reshape(torch.arange(batch_size), [batch_size, 1, 1])
    if flow.is_cuda:
        batch_range = batch_range.cuda()
    idx_batch_offset = (
        batch_range.repeat(1, input_height, input_width) * input_height * input_width
    )

    # Flatten everything
    coords_floor_flattened = coords_floor.reshape(-1, 2)
    coords_offset_flattened = coords_offset.reshape(-1, 2)
    idx_batch_offset_flattened = idx_batch_offset.reshape(-1)

    # Initialize results
    idxs_list = []
    weights_list = []

    # Loop over different di and dj to the four neighboring pixels
    for di in range(2):
        for dj in range(2):
            # Compute the neighboring pixel coordinates
            idxs_i = coords_floor_flattened[:, 0] + di
            idxs_j = coords_floor_flattened[:, 1] + dj
            # Compute the flat index into all pixels
            idxs = idx_batch_offset_flattened + idxs_j * input_width + idxs_i

            # Only count valid pixels
            mask = torch.nonzero(
                torch.logical_and(
                    torch.logical_and(idxs_i >= 0, idxs_i < input_width),
                    torch.logical_and(idxs_j >= 0, idxs_j < input_height),
                ),
                as_tuple=True,
            )
            valid_idxs = idxs[mask]
            valid_offsets = coords_offset_flattened[mask]

            # Compute weights according to bilinear interpolation
            weights_i = (1.0 - di) - (-1) ** di * valid_offsets[:, 0]
            weights_j = (1.0 - dj) - (-1) ** dj * valid_offsets[:, 1]
            weights = weights_i * weights_j

            # Append indices and weights
            idxs_list.append(valid_idxs)
            weights_list.append(weights)

    # Concatenate everything
    idxs = torch.cat(idxs_list, dim=0)
    weights = torch.cat(weights_list, dim=0)
    counts = torch.zeros(batch_size * input_width * input_height, dtype=weights.dtype)
    if flow.is_cuda:
        counts = counts.cuda()
    counts.scatter_add_(0, idxs, weights)
    range_map = counts.reshape(batch_size, 1, input_height, input_width)
    return range_map


def compute_fb_consistency(flow_ij, flow_ji):
    # Compare forward and backward flow
    flow_ji_in_i = warp(flow_ji, flow_ij)
    fb_sq_diff = torch.sum((flow_ij + flow_ji_in_i) ** 2, dim=1, keepdim=True)
    fb_sum_sq = torch.sum((flow_ij**2 + flow_ji_in_i**2), dim=1, keepdim=True)

    return fb_sq_diff, fb_sum_sq


def compute_occlusion(
    flow_ij,
    flow_ji,
    occlusion_estimation,
    occlusion_are_zeros=False,
    boundaries_occluded=True,
):
    """
    Compute occlusion mask.
    Args:
        flow_ij: the forward optical flow [B, 2, H, W]
        flow_ji: the backward optical flow [B, 2, H, W]
        occlusion_estimation: the algorithm chosed to estimate the occlusion. ('brox', 'fb_abs', 'wang')
        occlusion_are_zeros: the occluded regions are 0 (default false)
        boundaries_occluded: If True, treat flow vectors pointing off the boundaries
        as occluded. Otherwise explicitly mark them as unoccluded.
    Return:
        occlusion_mask: the estimated occlusion mask [B, 1, H, W]
    """

    # Compare forward and backward flow
    fb_sq_diff, fb_sum_sq = compute_fb_consistency(flow_ij, flow_ji)

    occlusion_mask = torch.zeros_like(flow_ij[:, :1, :, :])

    if occlusion_estimation == "none":
        B, _, H, W = flow_ij.shape
        occlusion_mask = torch.zeros(
            B, 1, H, W, dtype=flow_ij.dtype, device=flow_ij.device
        )
    elif occlusion_estimation == "brox":
        occlusion_mask = (fb_sq_diff > 0.01 * fb_sum_sq + 0.5).to(dtype=flow_ij.dtype)
    elif occlusion_estimation == "fb_abs":
        occlusion_mask = (fb_sq_diff**0.5 > 1.5).to(dtype=flow_ij.dtype)
    elif occlusion_estimation == "wang":
        range_map = compute_range_map(flow_ji)
        occlusion_mask = 1 - torch.clamp(range_map, min=0.0, max=1.0)

    if not boundaries_occluded:
        warp = flow_to_warp(flow_ij)
        occlusion_mask = torch.min(occlusion_mask, mask_invalid(warp))

    if occlusion_are_zeros:
        occlusion_mask = 1 - occlusion_mask
    return occlusion_mask


def photo_loss_fn(image1, warped_image, mask_level0):
    return (image1 - warped_image).abs() * mask_level0


def cal_unsup_loss(image1, image2, flow_forward, flow_backward, gamma):
    max_flow = MAX_FLOW
    n_predictions = len(flow_forward)
    flow_loss = 0.0

    for i in range(n_predictions):
        with torch.no_grad():
            occlusion_mask = compute_occlusion(
                flow_ij=flow_forward[i],
                flow_ji=flow_backward[-1],
                occlusion_estimation="wang",
                occlusion_are_zeros=True,
                boundaries_occluded=True,
            )
            coords = flow_to_warp(flow_forward[i])
            valid_warp_mask = mask_invalid(coords)
            mask_level0 = occlusion_mask * valid_warp_mask
        warped_image = warp(image2, flow_forward[i])
        photo_loss = photo_loss_fn(image1, warped_image, mask_level0)

        i_weight = gamma ** (n_predictions - i - 1)
        flow_loss += i_weight * photo_loss.mean()
    return flow_loss


def sequence_loss(flow_preds, flow_gt, valid, cfg):
    """Loss function defined over sequence of flow predictions"""

    gamma = cfg.gamma
    max_flow = cfg.max_flow
    n_predictions = len(flow_preds)
    flow_loss = 0.0
    flow_gt_thresholds = [5, 10, 20]

    # exlude invalid pixels and extremely large diplacements
    mag = torch.sum(flow_gt**2, dim=1).sqrt()
    valid = (valid >= 0.5) & (mag < max_flow)

    for i in range(n_predictions):
        i_weight = gamma ** (n_predictions - i - 1)
        i_loss = (flow_preds[i] - flow_gt).abs()
        flow_loss += i_weight * (valid[:, None] * i_loss).mean()

    epe = torch.sum((flow_preds[-1] - flow_gt) ** 2, dim=1).sqrt()
    epe = epe.view(-1)[valid.view(-1)]

    metrics = {
        "epe": epe.mean().item(),
        "1px": (epe < 1).float().mean().item(),
        "3px": (epe < 3).float().mean().item(),
        "5px": (epe < 5).float().mean().item(),
    }

    flow_gt_length = torch.sum(flow_gt**2, dim=1).sqrt()
    flow_gt_length = flow_gt_length.view(-1)[valid.view(-1)]
    for t in flow_gt_thresholds:
        e = epe[flow_gt_length < t]
        metrics.update({f"{t}-th-5px": (e < 5).float().mean().item()})

    return flow_loss, metrics



import torch

MAX_FLOW = 400


def sequence_loss(flow_preds, flow_gt, valid, cfg):
    """Loss function defined over sequence of flow predictions"""

    gamma = cfg.gamma
    max_flow = cfg.max_flow
    n_predictions = len(flow_preds)
    flow_loss = 0.0
    flow_gt_thresholds = [5, 10, 20]

    # exlude invalid pixels and extremely large diplacements
    mag = torch.sum(flow_gt**2, dim=1).sqrt()
    valid = (valid >= 0.5) & (mag < max_flow)

    for i in range(n_predictions):
        i_weight = gamma ** (n_predictions - i - 1)
        i_loss = (flow_preds[i] - flow_gt).abs()
        flow_loss += i_weight * (valid[:, None] * i_loss).mean()

    epe = torch.sum((flow_preds[-1] - flow_gt) ** 2, dim=1).sqrt()
    epe = epe.view(-1)[valid.view(-1)]

    metrics = {
        "epe": epe.mean().item(),
        "1px": (epe < 1).float().mean().item(),
        "3px": (epe < 3).float().mean().item(),
        "5px": (epe < 5).float().mean().item(),
    }

    flow_gt_length = torch.sum(flow_gt**2, dim=1).sqrt()
    flow_gt_length = flow_gt_length.view(-1)[valid.view(-1)]
    for t in flow_gt_thresholds:
        e = epe[flow_gt_length < t]
        metrics.update({f"{t}-th-5px": (e < 5).float().mean().item()})

    return flow_loss, metrics


def smooth_l1_loss(diff):
    cond = diff.abs() < 1
    loss = torch.where(cond, 0.5 * diff**2, diff.abs() - 0.5)
    return loss


def sequence_loss_smooth(flow_preds, flow_gt, valid, cfg):
    """Loss function defined over sequence of flow predictions"""

    gamma = cfg.gamma
    max_flow = cfg.max_flow
    n_predictions = len(flow_preds)
    flow_loss = 0.0
    flow_gt_thresholds = [5, 10, 20]

    # exlude invalid pixels and extremely large diplacements
    mag = torch.sum(flow_gt**2, dim=1).sqrt()
    valid = (valid >= 0.5) & (mag < max_flow)

    for i in range(n_predictions):
        i_weight = gamma ** (n_predictions - i - 1)
        i_loss = smooth_l1_loss((flow_preds[i] - flow_gt))
        flow_loss += i_weight * (valid[:, None] * i_loss).mean()

    epe = torch.sum((flow_preds[-1] - flow_gt) ** 2, dim=1).sqrt()
    epe = epe.view(-1)[valid.view(-1)]

    metrics = {
        "epe": epe.mean().item(),
        "1px": (epe < 1).float().mean().item(),
        "3px": (epe < 3).float().mean().item(),
        "5px": (epe < 5).float().mean().item(),
    }

    flow_gt_length = torch.sum(flow_gt**2, dim=1).sqrt()
    flow_gt_length = flow_gt_length.view(-1)[valid.view(-1)]
    for t in flow_gt_thresholds:
        e = epe[flow_gt_length < t]
        metrics.update({f"{t}-th-5px": (e < 5).float().mean().item()})

    return flow_loss, metrics



import torch
import math
import torch.nn as nn
import torch.nn.functional as F
from .attention import BroadMultiHeadAttention


class CrossAttentionLayer(nn.Module):
    def __init__(
        self,
        qk_dim,
        v_dim,
        query_token_dim,
        tgt_token_dim,
        num_heads=8,
        attn_drop=0.0,
        proj_drop=0.0,
        drop_path=0.0,
        dropout=0.0,
    ):
        super(CrossAttentionLayer, self).__init__()
        assert (
            qk_dim % num_heads == 0
        ), f"dim {qk_dim} should be divided by num_heads {num_heads}."
        assert (
            v_dim % num_heads == 0
        ), f"dim {v_dim} should be divided by num_heads {num_heads}."
        """
            Query Token:    [N, C]  -> [N, qk_dim]  (Q)
            Target Token:   [M, D]  -> [M, qk_dim]  (K),    [M, v_dim]  (V)
        """
        self.num_heads = num_heads
        head_dim = qk_dim // num_heads
        self.scale = head_dim**-0.5

        self.norm1 = nn.LayerNorm(query_token_dim)
        self.norm2 = nn.LayerNorm(query_token_dim)
        self.multi_head_attn = BroadMultiHeadAttention(qk_dim, num_heads)
        self.q, self.k, self.v = (
            nn.Linear(query_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, v_dim, bias=True),
        )

        self.proj = nn.Linear(v_dim, query_token_dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(query_token_dim, query_token_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(query_token_dim, query_token_dim),
            nn.Dropout(dropout),
        )

    def forward(self, query, tgt_token, size=None, ids_keep=None):
        """
        x: [BH1W1, H3W3, D]
        """

        if ids_keep is not None:
            tgt_token = torch.gather(
                tgt_token,
                dim=1,
                index=ids_keep.unsqueeze(-1).repeat(1, 1, tgt_token.shape[-1]),
            )

        short_cut = query
        query = self.norm1(query)

        q, k, v = self.q(query), self.k(tgt_token), self.v(tgt_token)

        x = self.multi_head_attn(q, k, v)

        x = short_cut + self.proj_drop(self.proj(x))

        x = x + self.drop_path(self.ffn(self.norm2(x)))

        return x


class CrossAttentionLayer_two_level(nn.Module):
    def __init__(
        self,
        qk_dim,
        v_dim,
        query_token_dim,
        tgt_token_dim,
        num_heads=8,
        attn_drop=0.0,
        proj_drop=0.0,
        drop_path=0.0,
        dropout=0.0,
    ):
        super(CrossAttentionLayer_two_level, self).__init__()
        assert (
            qk_dim % num_heads == 0
        ), f"dim {qk_dim} should be divided by num_heads {num_heads}."
        assert (
            v_dim % num_heads == 0
        ), f"dim {v_dim} should be divided by num_heads {num_heads}."
        """
            Query Token:    [N, C]  -> [N, qk_dim]  (Q)
            Target Token:   [M, D]  -> [M, qk_dim]  (K),    [M, v_dim]  (V)
        """
        self.num_heads = num_heads
        head_dim = qk_dim // num_heads
        self.scale = head_dim**-0.5

        self.norm1 = nn.LayerNorm(query_token_dim)
        self.norm2 = nn.LayerNorm(query_token_dim)
        self.multi_head_attn = BroadMultiHeadAttention(qk_dim, num_heads)
        self.q, self.k, self.v = (
            nn.Linear(query_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, v_dim, bias=True),
        )

        self.proj = nn.Linear(v_dim, query_token_dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(query_token_dim, query_token_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(query_token_dim, query_token_dim),
            nn.Dropout(dropout),
        )

    def compute_indices(self, size):
        H, W = size
        hs = [(0, int(H / 2)), (int(H / 2), H)]
        ws = [(0, int(W / 3)), (int(W / 3), int(W / 3) * 2), (int(W / 3) * 2, W)]
        return [(*h, *w) for h in hs for w in ws]

    def forward(self, query, tgt_token, size):
        """
        x: [BH1W1, H3W3, D]
        q ([1, 8, 128])  k torch.Size([5704, 48, 128]
        """
        short_cut = query
        query = self.norm1(query)

        q, k, v = self.q(query), self.k(tgt_token), self.v(tgt_token)
        B, HW, C = k.shape
        H, W = size

        res = []
        res.append(self.multi_head_attn(q[:, :2, :], k, v))

        k = k.reshape(B, H, W, C)
        v = v.reshape(B, H, W, C)

        indices = self.compute_indices(size)
        for idx, corners in enumerate(indices):
            h0, h1, w0, w1 = corners
            res.append(
                self.multi_head_attn(
                    q[:, 2 + idx : 3 + idx, :],
                    k[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                    v[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                )
            )

        x = torch.cat(res, dim=1)

        x = short_cut + self.proj_drop(self.proj(x))

        x = x + self.drop_path(self.ffn(self.norm2(x)))

        return x


class CrossAttentionLayer_convk3s2(nn.Module):
    def __init__(
        self,
        qk_dim,
        v_dim,
        query_token_dim,
        tgt_token_dim,
        num_heads=8,
        attn_drop=0.0,
        proj_drop=0.0,
        drop_path=0.0,
        dropout=0.0,
    ):
        super(CrossAttentionLayer_convk3s2, self).__init__()
        assert (
            qk_dim % num_heads == 0
        ), f"dim {qk_dim} should be divided by num_heads {num_heads}."
        assert (
            v_dim % num_heads == 0
        ), f"dim {v_dim} should be divided by num_heads {num_heads}."
        """
            Query Token:    [N, C]  -> [N, qk_dim]  (Q)
            Target Token:   [M, D]  -> [M, qk_dim]  (K),    [M, v_dim]  (V)
        """
        self.num_heads = num_heads
        head_dim = qk_dim // num_heads
        self.scale = head_dim**-0.5

        self.norm1 = nn.LayerNorm(query_token_dim)
        self.norm2 = nn.LayerNorm(query_token_dim)
        self.multi_head_attn = BroadMultiHeadAttention(qk_dim, num_heads)
        self.q, self.k, self.v = (
            nn.Linear(query_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, v_dim, bias=True),
        )

        self.proj = nn.Linear(v_dim, query_token_dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(query_token_dim, query_token_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(query_token_dim, query_token_dim),
            nn.Dropout(dropout),
        )

        # coarse level
        self.down_sample = nn.Conv2d(query_token_dim, query_token_dim, 3, 2, 1)
        self.norm1_coarse = nn.LayerNorm(query_token_dim)
        self.norm2_coarse = nn.LayerNorm(query_token_dim)
        self.q_coarse, self.k_coarse, self.v_coarse = (
            nn.Linear(query_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, v_dim, bias=True),
        )
        self.proj_coarse = nn.Linear(v_dim, query_token_dim)
        self.ffn_coarse = nn.Sequential(
            nn.Linear(query_token_dim, query_token_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(query_token_dim, query_token_dim),
            nn.Dropout(dropout),
        )

    def compute_indices_fine(self, size):
        H, W = size
        hs = [(0, int(H / 2)), (int(H / 2), H)]
        ws = [(0, int(W / 3)), (int(W / 3), int(W / 3) * 2), (int(W / 3) * 2, W)]
        return [(*h, *w) for h in hs for w in ws]

    def compute_indices_coarse(self, size):
        H, W = size
        hs = [(0, int(H / 2)), (int(H / 2), H)]
        ws = [(0, int(W / 2)), (int(W / 2), W)]
        return [(*h, *w) for h in hs for w in ws]

    def forward(self, query, tgt_token, size):
        """
        x: [BH1W1, H3W3, D]
        q ([1, 8, 128])  k torch.Size([5704, 48, 128]
        """
        assert query.shape[1] == 10, "Error: k3s2 version, query num must be 10"

        short_cut = query
        query = self.norm1(query[:, :6, :])

        # fine level
        q, k, v = self.q(query), self.k(tgt_token), self.v(tgt_token)
        B, HW, C = k.shape
        H, W = size

        res_fine = []

        k = k.reshape(B, H, W, C)
        v = v.reshape(B, H, W, C)

        indices = self.compute_indices_fine(size)
        for idx, corners in enumerate(indices):
            h0, h1, w0, w1 = corners
            res_fine.append(
                self.multi_head_attn(
                    q[:, idx : idx + 1, :],
                    k[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                    v[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                )
            )

        x = torch.cat(res_fine, dim=1)

        x = short_cut[:, :6, :] + self.proj_drop(self.proj(x))

        x_fine = x + self.drop_path(self.ffn(self.norm2(x)))

        # coarse_level
        tgt_token_coarse = tgt_token.reshape(B, H, W, -1).permute(0, 3, 1, 2)
        tgt_token_coarse = self.down_sample(tgt_token_coarse).permute(0, 2, 3, 1)
        H, W = tgt_token_coarse.shape[1:3]

        query = self.norm1_coarse(short_cut[:, 6:10, :])
        q, k, v = (
            self.q_coarse(query),
            self.k_coarse(tgt_token_coarse),
            self.v_coarse(tgt_token_coarse),
        )
        res_coarse = []

        indices = self.compute_indices_coarse((H, W))
        for idx, corners in enumerate(indices):
            h0, h1, w0, w1 = corners
            res_coarse.append(
                self.multi_head_attn(
                    q[:, idx : idx + 1, :],
                    k[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                    v[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                )
            )

        x = torch.cat(res_coarse, dim=1)

        x = short_cut[:, 6:10, :] + self.proj_drop(self.proj_coarse(x))

        x_coarse = x + self.drop_path(self.ffn_coarse(self.norm2_coarse(x)))

        x = torch.cat([x_fine, x_coarse], dim=1)

        return x


class CrossAttentionLayer_two_level_rep(nn.Module):
    def __init__(
        self,
        qk_dim,
        v_dim,
        query_token_dim,
        tgt_token_dim,
        num_heads=8,
        attn_drop=0.0,
        proj_drop=0.0,
        drop_path=0.0,
        dropout=0.0,
    ):
        super(CrossAttentionLayer_two_level_rep, self).__init__()
        assert (
            qk_dim % num_heads == 0
        ), f"dim {qk_dim} should be divided by num_heads {num_heads}."
        assert (
            v_dim % num_heads == 0
        ), f"dim {v_dim} should be divided by num_heads {num_heads}."
        """
            Query Token:    [N, C]  -> [N, qk_dim]  (Q)
            Target Token:   [M, D]  -> [M, qk_dim]  (K),    [M, v_dim]  (V)
        """
        self.num_heads = num_heads
        head_dim = qk_dim // num_heads
        self.scale = head_dim**-0.5

        self.norm1 = nn.LayerNorm(query_token_dim)
        self.norm2 = nn.LayerNorm(query_token_dim)
        self.multi_head_attn = BroadMultiHeadAttention(qk_dim, num_heads)
        self.q, self.k, self.v = (
            nn.Linear(query_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, v_dim, bias=True),
        )

        self.proj = nn.Linear(v_dim, query_token_dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(query_token_dim, query_token_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(query_token_dim, query_token_dim),
            nn.Dropout(dropout),
        )

    def compute_indices(self, size):
        H, W = size
        hs = [(0, int(H / 2)), (int(H / 2), H)]
        ws = [(0, int(W / 3)), (int(W / 3), int(W / 3) * 2), (int(W / 3) * 2, W)]
        return [(*h, *w) for h in hs for w in ws]

    def forward(self, query, tgt_token, size):
        """
        x: [BH1W1, H3W3, D]
        q ([1, 8, 128])  k torch.Size([5704, 48, 128]
        """
        short_cut = query
        query = self.norm1(query)

        q, k, v = self.q(query), self.k(tgt_token), self.v(tgt_token)
        B, HW, C = k.shape
        H, W = size

        res = []

        k = k.reshape(B, H, W, C)
        v = v.reshape(B, H, W, C)

        indices = self.compute_indices(size)

        h0, h1, w0, w1 = indices[0]
        res.append(
            self.multi_head_attn(
                q[:, 0:1, :],
                k[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                v[:, h0:h1, w0:w1, :].reshape(B, -1, C),
            )
        )
        h0, h1, w0, w1 = indices[1]
        res.append(
            self.multi_head_attn(
                q[:, 1:2, :],
                k[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                v[:, h0:h1, w0:w1, :].reshape(B, -1, C),
            )
        )

        for idx, corners in enumerate(indices):
            h0, h1, w0, w1 = corners
            res.append(
                self.multi_head_attn(
                    q[:, 2 + idx : 3 + idx, :],
                    k[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                    v[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                )
            )

        x = torch.cat(res, dim=1)

        x = short_cut + self.proj_drop(self.proj(x))

        x = x + self.drop_path(self.ffn(self.norm2(x)))

        return x


class CrossAttentionLayer_34(nn.Module):
    def __init__(
        self,
        qk_dim,
        v_dim,
        query_token_dim,
        tgt_token_dim,
        num_heads=8,
        attn_drop=0.0,
        proj_drop=0.0,
        drop_path=0.0,
        dropout=0.0,
    ):
        super(CrossAttentionLayer_34, self).__init__()
        assert (
            qk_dim % num_heads == 0
        ), f"dim {qk_dim} should be divided by num_heads {num_heads}."
        assert (
            v_dim % num_heads == 0
        ), f"dim {v_dim} should be divided by num_heads {num_heads}."
        """
            Query Token:    [N, C]  -> [N, qk_dim]  (Q)
            Target Token:   [M, D]  -> [M, qk_dim]  (K),    [M, v_dim]  (V)
        """
        self.num_heads = num_heads
        head_dim = qk_dim // num_heads
        self.scale = head_dim**-0.5

        self.norm1 = nn.LayerNorm(query_token_dim)
        self.norm2 = nn.LayerNorm(query_token_dim)
        self.multi_head_attn = BroadMultiHeadAttention(qk_dim, num_heads)
        self.q, self.k, self.v = (
            nn.Linear(query_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, v_dim, bias=True),
        )

        self.proj = nn.Linear(v_dim, query_token_dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(query_token_dim, query_token_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(query_token_dim, query_token_dim),
            nn.Dropout(dropout),
        )

    def compute_indices(self, size):
        H, W = size
        hs = [(0, int(H / 3)), (int(H / 3), int(H / 3) * 2), (int(H / 3) * 2, H)]
        ws = [
            (0, int(W / 4)),
            (int(W / 4), int(W / 4) * 2),
            (int(W / 4) * 2, int(W / 4) * 3),
            (int(W / 4) * 3, W),
        ]
        return [(*h, *w) for h in hs for w in ws]

    def forward(self, query, tgt_token, size):
        """
        x: [BH1W1, H3W3, D]
        q ([1, 8, 128])  k torch.Size([5704, 48, 128]
        """
        short_cut = query
        query = self.norm1(query)

        q, k, v = self.q(query), self.k(tgt_token), self.v(tgt_token)
        B, HW, C = k.shape
        H, W = size

        res = []

        k = k.reshape(B, H, W, C)
        v = v.reshape(B, H, W, C)

        indices = self.compute_indices(size)
        for idx, corners in enumerate(indices):
            h0, h1, w0, w1 = corners
            res.append(
                self.multi_head_attn(
                    q[:, idx : idx + 1, :],
                    k[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                    v[:, h0:h1, w0:w1, :].reshape(B, -1, C),
                )
            )

        x = torch.cat(res, dim=1)

        x = short_cut + self.proj_drop(self.proj(x))

        x = x + self.drop_path(self.ffn(self.norm2(x)))

        return x



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import einsum

from einops.layers.torch import Rearrange
from einops import rearrange


class BroadMultiHeadAttention(nn.Module):
    def __init__(self, dim, heads):
        super(BroadMultiHeadAttention, self).__init__()
        self.dim = dim
        self.heads = heads
        self.scale = (dim / heads) ** -0.5
        self.attend = nn.Softmax(dim=-1)

    def attend_with_rpe(self, Q, K):
        if Q.shape[0] == 1:
            Q = rearrange(Q.squeeze(0), "i (heads d) -> heads i d", heads=self.heads)
            K = rearrange(K, "b j (heads d) -> b heads j d", heads=self.heads)

            dots = (
                einsum("hid, bhjd -> bhij", Q, K) * self.scale
            )  # (b hw) heads 1 pointnum
        else:
            Q = rearrange(Q, "b i (heads d) -> b heads i d", heads=self.heads)
            K = rearrange(K, "b j (heads d) -> b heads j d", heads=self.heads)

            dots = (
                einsum("bhid, bhjd -> bhij", Q, K) * self.scale
            )  # (b hw) heads 1 pointnum

        return self.attend(dots)

    def forward(self, Q, K, V):
        attn = self.attend_with_rpe(Q, K)
        B, _, _ = K.shape
        _, N, _ = Q.shape

        V = rearrange(V, "b j (heads d) -> b heads j d", heads=self.heads)

        out = einsum("bhij, bhjd -> bhid", attn, V)
        out = rearrange(out, "b heads n d -> b n (heads d)", b=B, n=N)

        return out


class MultiHeadAttention(nn.Module):
    def __init__(self, dim, heads):
        super(MultiHeadAttention, self).__init__()
        self.dim = dim
        self.heads = heads
        self.scale = (dim / heads) ** -0.5
        self.attend = nn.Softmax(dim=-1)

    def attend_with_rpe(self, Q, K):
        Q = rearrange(Q, "b i (heads d) -> b heads i d", heads=self.heads)
        K = rearrange(K, "b j (heads d) -> b heads j d", heads=self.heads)

        dots = (
            einsum("bhid, bhjd -> bhij", Q, K) * self.scale
        )  # (b hw) heads 1 pointnum

        return self.attend(dots)

    def forward(self, Q, K, V):
        attn = self.attend_with_rpe(Q, K)
        B, HW, _ = Q.shape

        V = rearrange(V, "b j (heads d) -> b heads j d", heads=self.heads)

        out = einsum("bhij, bhjd -> bhid", attn, V)
        out = rearrange(out, "b heads hw d -> b hw (heads d)", b=B, hw=HW)

        return out


# class MultiHeadAttentionRelative_encoder(nn.Module):
#     def __init__(self, dim, heads):
#         super(MultiHeadAttentionRelative, self).__init__()
#         self.dim = dim
#         self.heads = heads
#         self.scale = (dim/heads) ** -0.5
#         self.attend = nn.Softmax(dim=-1)

#     def attend_with_rpe(self, Q, K, Q_r, K_r):
#         """
#             Q: [BH1W1, H3W3, dim]
#             K: [BH1W1, H3W3, dim]
#             Q_r: [BH1W1, H3W3, H3W3, dim]
#             K_r: [BH1W1, H3W3, H3W3, dim]
#         """

#         Q = rearrange(Q, 'b i (heads d) -> b heads i d', heads=self.heads) # [BH1W1, heads, H3W3, dim]
#         K = rearrange(K, 'b j (heads d) -> b heads j d', heads=self.heads) # [BH1W1, heads, H3W3, dim]
#         K_r = rearrange(K_r, 'b j (heads d) -> b heads j d', heads=self.heads) # [BH1W1, heads, H3W3, dim]
#         Q_r = rearrange(Q_r, 'b j (heads d) -> b heads j d', heads=self.heads) # [BH1W1, heads, H3W3, dim]

#         # context-context similarity
#         c_c = einsum('bhid, bhjd -> bhij', Q, K) * self.scale # [(B H1W1) heads H3W3 H3W3]
#         # context-position similarity
#         c_p = einsum('bhid, bhjd -> bhij', Q, K_r) * self.scale # [(B H1W1) heads 1 H3W3]
#         # position-context similarity
#         p_c = einsum('bhijd, bhikd -> bhijk', Q_r[:,:,:,None,:], K[:,:,:,None,:])
#         p_c = torch.squeeze(p_c, dim=4)
#         p_c = p_c.permute(0, 1, 3, 2)
#         dots = c_c + c_p + p_c
#         return self.attend(dots)

#     def forward(self, Q, K, V, Q_r, K_r):
#         attn = self.attend_with_rpe(Q, K, Q_r, K_r)
#         B, HW, _ = Q.shape

#         V = rearrange(V, 'b j (heads d) -> b heads j d', heads=self.heads)

#         out = einsum('bhij, bhjd -> bhid', attn, V)
#         out = rearrange(out, 'b heads hw d -> b hw (heads d)', b=B, hw=HW)

#         return out


class MultiHeadAttentionRelative(nn.Module):
    def __init__(self, dim, heads):
        super(MultiHeadAttentionRelative, self).__init__()
        self.dim = dim
        self.heads = heads
        self.scale = (dim / heads) ** -0.5
        self.attend = nn.Softmax(dim=-1)

    def attend_with_rpe(self, Q, K, Q_r, K_r):
        """
        Q: [BH1W1, 1, dim]
        K: [BH1W1, H3W3, dim]
        Q_r: [BH1W1, H3W3, dim]
        K_r: [BH1W1, H3W3, dim]
        """

        Q = rearrange(
            Q, "b i (heads d) -> b heads i d", heads=self.heads
        )  # [BH1W1, heads, 1, dim]
        K = rearrange(
            K, "b j (heads d) -> b heads j d", heads=self.heads
        )  # [BH1W1, heads, H3W3, dim]
        K_r = rearrange(
            K_r, "b j (heads d) -> b heads j d", heads=self.heads
        )  # [BH1W1, heads, H3W3, dim]
        Q_r = rearrange(
            Q_r, "b j (heads d) -> b heads j d", heads=self.heads
        )  # [BH1W1, heads, H3W3, dim]

        # context-context similarity
        c_c = einsum("bhid, bhjd -> bhij", Q, K) * self.scale  # [(B H1W1) heads 1 H3W3]
        # context-position similarity
        c_p = (
            einsum("bhid, bhjd -> bhij", Q, K_r) * self.scale
        )  # [(B H1W1) heads 1 H3W3]
        # position-context similarity
        p_c = (
            einsum("bhijd, bhikd -> bhijk", Q_r[:, :, :, None, :], K[:, :, :, None, :])
            * self.scale
        )
        p_c = torch.squeeze(p_c, dim=4)
        p_c = p_c.permute(0, 1, 3, 2)
        dots = c_c + c_p + p_c
        return self.attend(dots)

    def forward(self, Q, K, V, Q_r, K_r):
        attn = self.attend_with_rpe(Q, K, Q_r, K_r)
        B, HW, _ = Q.shape

        V = rearrange(V, "b j (heads d) -> b heads j d", heads=self.heads)

        out = einsum("bhij, bhjd -> bhid", attn, V)
        out = rearrange(out, "b heads hw d -> b hw (heads d)", b=B, hw=HW)

        return out


def LinearPositionEmbeddingSine(x, dim=128, NORMALIZE_FACOR=1 / 200):
    # 200 should be enough for a 8x downsampled image
    # assume x to be [_, _, 2]
    # print(NORMALIZE_FACOR)
    freq_bands = torch.linspace(
        0, dim // 4 - 1, dim // 4, dtype=x.dtype, device=x.device
    )
    return torch.cat(
        [
            torch.sin(3.14 * x[..., -2:-1] * freq_bands * NORMALIZE_FACOR),
            torch.cos(3.14 * x[..., -2:-1] * freq_bands * NORMALIZE_FACOR),
            torch.sin(3.14 * x[..., -1:] * freq_bands * NORMALIZE_FACOR),
            torch.cos(3.14 * x[..., -1:] * freq_bands * NORMALIZE_FACOR),
        ],
        dim=-1,
    )


def ExpPositionEmbeddingSine(x, dim=128, NORMALIZE_FACOR=1 / 200):
    # 200 should be enough for a 8x downsampled image
    # assume x to be [_, _, 2]
    freq_bands = torch.linspace(
        0, dim // 4 - 1, dim // 4, dtype=x.dtype, device=x.device
    )
    return torch.cat(
        [
            torch.sin(x[..., -2:-1] * (NORMALIZE_FACOR * 2**freq_bands)),
            torch.cos(x[..., -2:-1] * (NORMALIZE_FACOR * 2**freq_bands)),
            torch.sin(x[..., -1:] * (NORMALIZE_FACOR * 2**freq_bands)),
            torch.cos(x[..., -1:] * (NORMALIZE_FACOR * 2**freq_bands)),
        ],
        dim=-1,
    )



import torch
import torch.nn as nn
from torch.autograd import Variable
import torch.nn.functional as F


class MultiScale_UP(nn.Module):
    def __init__(
        self,
        loss_type="L1",
        weight=[1.0, 0.5, 0.25],
        valid_range=None,
        removezero=False,
        use_valid_range=False,
    ):
        super().__init__()
        self.loss_type = loss_type
        self.weight = weight
        self.valid_range = valid_range
        self.removezero = removezero
        self.use_valid_range = use_valid_range

    def forward(self, preds, inputs):
        # Multi scale Loss, where flow upsampling is automatically computed here
        loss = 0
        loss_list = []

        target = inputs["flows"][:, 0]
        output = preds["flow_preds"]
        extra_mask = inputs.get("extra_mask")
        b, _, h, w = target.size()

        for i, cur_output in enumerate(output):
            # Compute loss for each level
            realflow = F.interpolate(
                cur_output, (h, w), mode="bilinear", align_corners=True
            )
            realflow[:, 0, :, :] = realflow[:, 0, :, :] * (w / cur_output.shape[3])
            realflow[:, 1, :, :] = realflow[:, 1, :, :] * (h / cur_output.shape[2])

            with torch.no_grad():
                if i == 0:
                    epe = realEPE(realflow, target, extra_mask=extra_mask)

            if self.loss_type == "L2":
                lossvalue = torch.norm(realflow - target, p=2, dim=1)
            elif self.loss_type == "robust":
                lossvalue = (realflow - target).abs().sum(dim=1) + 1e-8
                lossvalue = lossvalue**0.4
            elif self.loss_type == "L1":
                lossvalue = (realflow - target).abs().sum(dim=1)
            else:
                raise NotImplementedError

            if self.use_valid_range and self.valid_range is not None:
                # Filter out the pixels whose gt flow is out of our search range
                with torch.no_grad():
                    mask = (target[:, 0, :, :].abs() <= self.valid_range[i][1]) & (
                        target[:, 1, :, :].abs() <= self.valid_range[i][0]
                    )
            else:
                with torch.no_grad():
                    mask = torch.ones(target[:, 0, :, :].shape).type_as(target)

            lossvalue = lossvalue * mask.float()

            if extra_mask is not None:
                val = extra_mask > 0
                lossvalue = lossvalue[val]
                cur_loss = lossvalue.mean() * self.weight[i]
                assert lossvalue.shape[0] == extra_mask.sum()
            else:
                cur_loss = lossvalue.mean() * self.weight[i]

            loss += cur_loss
            loss_list.append(cur_loss)

        loss = loss / len(output)
        return {"loss": loss, "loss_list": loss_list, "epe": epe}


#################################################


def random_select_points(x, y, x_, y_, samples=10):
    idx = torch.randperm(x.shape[0])
    x = x[idx[:samples], :]
    y = y[idx[:samples], :]
    x_ = x_[idx[:samples], :]
    y_ = y_[idx[:samples], :]
    return x, y, x_, y_


def subspace_loss_batch(flow):
    # https://openaccess.thecvf.com/content_CVPR_2019/papers/Zhong_Unsupervised_Deep_Epipolar_Flow_for_Stationary_or_Dynamic_Scenes_CVPR_2019_paper.pdf
    B, _, H, W = flow.size()
    xx = Variable(torch.arange(0, W).view(1, -1).repeat(H, 1).cuda())
    yy = Variable(torch.arange(0, H).view(-1, 1).repeat(1, W).cuda())
    grid_x = xx.view(1, 1, H, W).repeat(B, 1, 1, 1).float()
    grid_y = yy.view(1, 1, H, W).repeat(B, 1, 1, 1).float()

    flow_u = flow[:, 0, :, :].unsqueeze(1)
    flow_v = flow[:, 1, :, :].unsqueeze(1)

    pos_x = grid_x + flow_u
    pos_y = grid_y + flow_v

    inside_x = (pos_x <= (W - 1)) & (pos_x >= 0.0)
    inside_y = (pos_y <= (H - 1)) & (pos_y >= 0.0)

    inside = inside_x & inside_y

    loss = 0

    least_num = 2000

    list_X = []
    list_X_ = []
    for i in range(B):
        grid_x_i = grid_x[i, :, :, :]
        grid_y_i = grid_y[i, :, :, :]
        pos_x_i = pos_x[i, :, :, :]
        pos_y_i = pos_y[i, :, :, :]
        inside_i = inside[i, :, :, :]

        if inside_i.sum() > least_num:
            x = torch.masked_select(grid_x_i, inside_i).view(-1, 1)
            y = torch.masked_select(grid_y_i, inside_i).view(-1, 1)
            x_ = torch.masked_select(pos_x_i, inside_i).view(-1, 1)
            y_ = torch.masked_select(pos_y_i, inside_i).view(-1, 1)
            x, y, x_, y_ = random_select_points(x, y, x_, y_, samples=least_num)
            o = torch.ones_like(x)
            x, y, x_, y_ = x / W, y / W, x_ / W, y_ / W
            X = torch.cat((x, x, x, y, y, y, o, o, o), 1).permute(1, 0)
            X_ = torch.cat((x_, y_, o, x_, y_, o, x_, y_, o), 1).permute(1, 0)
            list_X.append(X.unsqueeze(0))
            list_X_.append(X_.unsqueeze(0))

    all_X = torch.cat(list_X)
    all_X_ = torch.cat(list_X_)
    M = all_X * all_X_
    lambda1 = 10
    MTM = lambda1 * torch.matmul(M.permute(0, 2, 1), M)
    I = torch.eye(MTM.size()[1]).type_as(MTM).unsqueeze(0).repeat(B, 1, 1)
    MTM_inverse = torch.inverse((I + MTM))
    C = torch.matmul(MTM_inverse, MTM)
    C2 = C**2
    loss1 = torch.sum(C2.view(-1, 1), dim=0)

    loss2 = lambda1 * torch.sum(((torch.matmul(M, C) - M) ** 2).view(-1, 1), dim=0)

    loss += loss1 + loss2

    return loss / B


def EPE_flow(input_flow, target_flow):
    return torch.norm(target_flow - input_flow, p=2, dim=1).mean()


class L1(nn.Module):
    def __init__(self):
        super(L1, self).__init__()

    def forward(self, output, target):
        lossvalue = torch.abs(output - target).mean()
        return lossvalue


class L2(nn.Module):
    def __init__(self):
        super(L2, self).__init__()

    def forward(self, output, target):
        lossvalue = torch.norm(output - target, p=2, dim=1).mean()
        return lossvalue


class L1Loss(nn.Module):
    def __init__(self):
        super(L1Loss, self).__init__()
        self.loss = L1()
        self.loss_labels = ["L1", "EPE"]

    def forward(self, output, target):
        lossvalue = self.loss(output, target)
        epevalue = EPE_flow(output, target)
        return [lossvalue, epevalue]


class L2Loss(nn.Module):
    def __init__(self):
        super(L2Loss, self).__init__()
        self.loss = L2()
        self.loss_labels = ["L2", "EPE"]

    def forward(self, output, target):
        lossvalue = self.loss(output, target)
        epevalue = EPE_flow(output, target)
        return [lossvalue, epevalue]


########################################################


def warp(x, flo):
    """
    warp an image/tensor (im2) back to im1, according to the optical flow
    x: [B, C, H, W] (im2)
    flo: [B, 2, H, W] flow
    """
    B, C, H, W = x.size()
    # mesh grid
    xx = torch.arange(0, W).view(1, -1).repeat(H, 1)
    yy = torch.arange(0, H).view(-1, 1).repeat(1, W)
    xx = xx.view(1, 1, H, W).repeat(B, 1, 1, 1)
    yy = yy.view(1, 1, H, W).repeat(B, 1, 1, 1)
    grid = torch.cat((xx, yy), 1).float()

    if x.is_cuda:
        grid = grid.cuda()
    vgrid = Variable(grid) + flo

    # scale grid to [-1,1]
    vgrid[:, 0, :, :] = 2.0 * vgrid[:, 0, :, :] / max(W - 1, 1) - 1.0
    vgrid[:, 1, :, :] = 2.0 * vgrid[:, 1, :, :] / max(H - 1, 1) - 1.0

    vgrid = vgrid.permute(0, 2, 3, 1)
    output = nn.functional.grid_sample(x, vgrid)
    mask = torch.autograd.Variable(torch.ones(x.size())).cuda()
    mask = nn.functional.grid_sample(mask, vgrid)

    mask[mask < 0.9999] = 0
    mask[mask > 0] = 1

    return output * mask


def warp_(self, tensorInput, tensorFlow):
    if (
        hasattr(self, "tensorPartial") == False
        or self.tensorPartial.size(0) != tensorFlow.size(0)
        or self.tensorPartial.size(2) != tensorFlow.size(2)
        or self.tensorPartial.size(3) != tensorFlow.size(3)
    ):
        self.tensorPartial = tensorFlow.new_ones(
            tensorFlow.size(0), 1, tensorFlow.size(2), tensorFlow.size(3)
        )
        # end

    if (
        hasattr(self, "tensorGrid") == False
        or self.tensorGrid.size(0) != tensorFlow.size(0)
        or self.tensorGrid.size(2) != tensorFlow.size(2)
        or self.tensorGrid.size(3) != tensorFlow.size(3)
    ):
        tensorHorizontal = (
            torch.linspace(-1.0, 1.0, tensorFlow.size(3))
            .view(1, 1, 1, tensorFlow.size(3))
            .expand(tensorFlow.size(0), -1, tensorFlow.size(2), -1)
        )
        tensorVertical = (
            torch.linspace(-1.0, 1.0, tensorFlow.size(2))
            .view(1, 1, tensorFlow.size(2), 1)
            .expand(tensorFlow.size(0), -1, -1, tensorFlow.size(3))
        )

        self.tensorGrid = torch.cat([tensorHorizontal, tensorVertical], 1).cuda()
        # end

    tensorInput = torch.cat([tensorInput, self.tensorPartial], 1)
    tensorFlow = torch.cat(
        [
            tensorFlow[:, 0:1, :, :] / ((tensorInput.size(3) - 1.0) / 2.0),
            tensorFlow[:, 1:2, :, :] / ((tensorInput.size(2) - 1.0) / 2.0),
        ],
        1,
    )

    tensorOutput = torch.nn.functional.grid_sample(
        input=tensorInput,
        grid=(self.tensorGrid + tensorFlow).permute(0, 2, 3, 1),
        mode="bilinear",
        padding_mode="zeros",
    )

    tensorMask = tensorOutput[:, -1:, :, :]
    tensorMask[tensorMask > 0.999] = 1.0
    tensorMask[tensorMask < 1.0] = 0.0

    return tensorOutput[:, :-1, :, :] * tensorMask


def EPE(input_flow, target_flow, sparse=False, mean=True, extra_mask=None):
    EPE_map = torch.norm(target_flow - input_flow, 2, 1)
    batch_size = EPE_map.size(0)

    if sparse:
        # invalid flow is defined with both flow coordinates to be exactly 0
        mask = (target_flow[:, 0] == 0) & (target_flow[:, 1] == 0)
        EPE_map = EPE_map[~mask]

    if extra_mask is not None:
        EPE_map = EPE_map[extra_mask.byte()]

    if mean:
        return EPE_map.mean()
    else:
        return EPE_map.sum() / batch_size


def realEPE(
    output,
    target,
    sparse=False,
    valid_range=None,
    extra_mask=None,
    use_valid_range=False,
):
    b, _, h, w = target.size()
    upsampled_output = output
    if use_valid_range and valid_range is not None:
        mask = (target[:, 0, :, :].abs() <= valid_range[1]) & (
            target[:, 1, :, :].abs() <= valid_range[0]
        )
        mask = mask.unsqueeze(1).expand(-1, 2, -1, -1).float()
        upsampled_output = upsampled_output * mask
        target = target * mask
    return EPE(upsampled_output, target, sparse, mean=True, extra_mask=extra_mask)



import configparser

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from ptlflow.utils.utils import forward_interpolate_batch
from .update import BasicUpdateBlock
from .extractor import BasicEncoder, BasicConvEncoder
from .corr import CorrBlock
from .utils.utils import coords_grid, upflow8
from .swin_transformer import POLAUpdate, MixAxialPOLAUpdate
from .loss import compute_supervision_coarse, compute_coarse_loss, backwarp
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float, use_matching_loss: bool):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow
        self.use_matching_loss = use_matching_loss

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        soft_corr_map = outputs["soft_corr_map"]
        image1 = inputs["images"][:, 0]
        image2 = inputs["images"][:, 1]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        # original RAFT loss
        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exclude invalid pixels and extremely large displacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid[:, None].float() * i_loss).mean()

        if self.use_matching_loss:
            # enable global matching loss. Try to use it in late stages of the trianing
            img_2back1 = backwarp(image2, flow_gt)
            occlusionMap = (image1 - img_2back1).mean(1, keepdims=True)  # (N, H, W)
            occlusionMap = torch.abs(occlusionMap) > 20
            occlusionMap = occlusionMap.float()

            conf_matrix_gt = compute_supervision_coarse(
                flow_gt, occlusionMap, 8
            )  # 8 from RAFT downsample

            matchLossCfg = configparser.ConfigParser()
            matchLossCfg.POS_WEIGHT = 1
            matchLossCfg.NEG_WEIGHT = 1
            matchLossCfg.FOCAL_ALPHA = 0.25
            matchLossCfg.FOCAL_GAMMA = 2.0
            matchLossCfg.COARSE_TYPE = "cross_entropy"
            match_loss = compute_coarse_loss(
                soft_corr_map, conf_matrix_gt, matchLossCfg
            )

            flow_loss = flow_loss + 0.01 * match_loss

        return flow_loss


class GMFlowNet(BaseModel):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflownet-things-9f061ac7.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflownet-kitti-712b4660.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0.0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        use_matching_loss: bool = False,
        use_mix_attn: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8,
            loss_fn=SequenceLoss(gamma, max_flow, use_matching_loss),
            **kwargs,
        )

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.dropout = dropout
        self.gamma = gamma
        self.max_flow = max_flow
        self.iters = iters
        self.use_matching_loss = use_matching_loss
        self.use_mix_attn = use_mix_attn

        self.hidden_dim = hdim = 128
        self.context_dim = cdim = 128

        # feature network, context network, and update block
        if self.use_mix_attn:
            self.fnet = nn.Sequential(
                BasicConvEncoder(output_dim=256, norm_fn="instance", dropout=dropout),
                MixAxialPOLAUpdate(embed_dim=256, depth=6, num_head=8, window_size=7),
            )
        else:
            self.fnet = nn.Sequential(
                BasicConvEncoder(output_dim=256, norm_fn="instance", dropout=dropout),
                POLAUpdate(
                    embed_dim=256, depth=6, num_head=8, window_size=7, neig_win_num=1
                ),
            )

        self.cnet = BasicEncoder(
            output_dim=hdim + cdim, norm_fn="batch", dropout=dropout
        )
        self.update_block = BasicUpdateBlock(
            corr_levels=corr_levels,
            corr_radius=corr_radius,
            hidden_dim=hdim,
            input_dim=cdim,
        )

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)
        coords1 = coords_grid(N, H // 8, W // 8, dtype=img.dtype, device=img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the feature network
        fmap1, fmap2 = self.fnet([image1, image2])

        # # Self-attention update
        # fmap1 = self.transEncoder(fmap1)
        # fmap2 = self.transEncoder(fmap2)

        corr_fn = CorrBlock(fmap1, fmap2, radius=self.corr_radius)

        # run the context network
        cnet = self.cnet(image1)
        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)

        coords0, coords1 = self.initialize_flow(image1)

        # Correlation as initialization
        N, fC, fH, fW = fmap1.shape
        corrMap = corr_fn.corrMap

        # _, coords_index = torch.max(corrMap, dim=-1) # no gradient here
        softCorrMap = F.softmax(corrMap, dim=2) * F.softmax(
            corrMap, dim=1
        )  # (N, fH*fW, fH*fW)

        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            forward_flow = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])
            coords1 = coords1 + forward_flow
        else:
            # print('matching as init')
            # mutual match selection
            match12, match_idx12 = softCorrMap.max(dim=2)  # (N, fH*fW)
            match21, match_idx21 = softCorrMap.max(dim=1)

            for b_idx in range(N):
                match21_b = match21[b_idx, :]
                match_idx12_b = match_idx12[b_idx, :]
                match21[b_idx, :] = match21_b[match_idx12_b]

            matched = (match12 - match21) == 0  # (N, fH*fW)
            coords_index = (
                torch.arange(fH * fW).unsqueeze(0).repeat(N, 1).to(softCorrMap.device)
            )
            coords_index[matched] = match_idx12[matched]

            # matched coords
            coords_index = coords_index.reshape(N, fH, fW)
            coords_x = coords_index % fW
            coords_y = coords_index // fW

            coords_xy = torch.stack([coords_x, coords_y], dim=1).to(dtype=coords1.dtype)
            coords1 = coords_xy

        # Iterative update
        flow_predictions = []
        for itr in range(self.iters):
            coords1 = coords1.detach()
            corr = corr_fn(coords1)  # index correlation volume

            flow = coords1 - coords0
            net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow8(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)

            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            flow_predictions.append(flow_up)

        if self.training:
            outputs = {
                "flows": flow_up[:, None],
                "flow_preds": flow_predictions,
                "soft_corr_map": softCorrMap,
            }
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": coords1 - coords0}

        return outputs


class GMFlowNetMix(GMFlowNet):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflownet_mix-things-8396f0a1.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/gmflownet_mix-sintel-33492618.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        use_matching_loss: bool = False,
        use_mix_attn: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            corr_levels,
            corr_radius,
            dropout,
            gamma,
            max_flow,
            iters,
            use_matching_loss,
            use_mix_attn,
            **kwargs,
        )


@register_model
@trainable
class gmflownet(GMFlowNet):
    pass


@register_model
@trainable
class gmflownet_mix(GMFlowNetMix):
    pass



import torch
import torch.nn.functional as F
from .utils import bilinear_sampler


class CorrBlock:
    def __init__(self, fmap1, fmap2, gamma, corr2, num_levels=4, radius=4):
        self.num_levels = num_levels
        self.radius = radius
        self.corr_pyramid = []

        # all pairs correlation
        corr = CorrBlock.corr(fmap1, fmap2)
        corr = corr + gamma * corr2

        batch, h1, w1, dim, h2, w2 = corr.shape
        corr = corr.reshape(batch * h1 * w1, dim, h2, w2)

        self.corr_pyramid.append(corr)
        for i in range(self.num_levels - 1):
            corr = F.avg_pool2d(corr, 2, stride=2)
            self.corr_pyramid.append(corr)

    def __call__(self, coords):
        r = self.radius
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        out_pyramid = []
        for i in range(self.num_levels):
            corr = self.corr_pyramid[i]
            dx = torch.linspace(
                -r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device
            )
            dy = torch.linspace(
                -r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device
            )
            delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1)

            centroid_lvl = coords.reshape(batch * h1 * w1, 1, 1, 2) / 2**i
            delta_lvl = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
            coords_lvl = centroid_lvl + delta_lvl

            corr = bilinear_sampler(corr, coords_lvl)
            corr = corr.view(batch, h1, w1, -1)
            out_pyramid.append(corr)

        out = torch.cat(out_pyramid, dim=-1)
        return out.permute(0, 3, 1, 2).contiguous()

    @staticmethod
    def corr(fmap1, fmap2):
        batch, dim, ht, wd = fmap1.shape
        fmap1 = fmap1.view(batch, dim, ht * wd)
        fmap2 = fmap2.view(batch, dim, ht * wd)

        corr = torch.matmul(fmap1.transpose(1, 2), fmap2)
        corr = corr.view(batch, ht, wd, 1, ht, wd)
        return corr / torch.sqrt(torch.tensor(dim))



from typing import Optional, Sequence

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )
import torch
import torch.nn as nn

from ptlflow.utils.registry import register_model, trainable
from .pwc_modules import upsample2d_as, initialize_msra
from .pwc_modules import (
    WarpingLayer,
    FeatureExtractor,
    ContextNetwork,
    FlowEstimatorDense,
)
from ..base_model.base_model import BaseModel
from .losses import MultiScaleEPE_PWC


class IRRPWCNet(BaseModel):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/irr_pwcnet-things-3f7fb8ca.ckpt"
    }

    def __init__(
        self,
        div_flow: float = 0.05,
        search_range: int = 4,
        output_level: int = 4,
        num_chs: Sequence[int] = (3, 16, 32, 64, 96, 128, 196),
        train_batch_size: Optional[int] = None,
        **kwargs,
    ):
        super(IRRPWCNet, self).__init__(
            output_stride=64,
            loss_fn=MultiScaleEPE_PWC(
                train_batch_size=train_batch_size, div_flow=div_flow
            ),
            **kwargs,
        )

        self.div_flow = div_flow
        self.search_range = search_range
        self.output_level = output_level
        self.num_chs = num_chs
        self.train_batch_size = train_batch_size

        self.leakyRELU = nn.LeakyReLU(0.1, inplace=True)

        self.feature_pyramid_extractor = FeatureExtractor(self.num_chs)
        self.warping_layer = WarpingLayer()

        self.flow_estimators = nn.ModuleList()
        self.dim_corr = (self.search_range * 2 + 1) ** 2
        for l, ch in enumerate(self.num_chs[::-1]):
            if l > self.output_level:
                break

            if l == 0:
                num_ch_in = self.dim_corr
            else:
                num_ch_in = self.dim_corr + ch + 2

            layer = FlowEstimatorDense(num_ch_in)
            self.flow_estimators.append(layer)

        self.context_networks = ContextNetwork(self.dim_corr + 32 + 2 + 448 + 2)

        self.corr = SpatialCorrelationSampler(
            kernel_size=1, patch_size=2 * self.search_range + 1, padding=0
        )

        initialize_msra(self.modules())

    def forward(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=0.0,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=False,
        )

        x1_raw = images[:, 0]
        x2_raw = images[:, 1]
        _, _, height_im, width_im = x1_raw.size()

        # on the bottom level are original images
        x1_pyramid = self.feature_pyramid_extractor(x1_raw) + [x1_raw]
        x2_pyramid = self.feature_pyramid_extractor(x2_raw) + [x2_raw]

        # outputs
        output_dict = {}
        flows = []

        # init
        (
            b_size,
            _,
            h_x1,
            w_x1,
        ) = x1_pyramid[0].size()
        init_dtype = x1_pyramid[0].dtype
        init_device = x1_pyramid[0].device
        flow = torch.zeros(b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device)

        for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):
            # warping
            if l == 0:
                x2_warp = x2
            else:
                flow = upsample2d_as(flow, x1, mode="bilinear")
                x2_warp = self.warping_layer(
                    x2, flow, height_im, width_im, self.div_flow
                )

            # correlation
            out_corr = self.corr(x1, x2_warp)
            out_corr = out_corr.view(
                out_corr.shape[0], -1, out_corr.shape[3], out_corr.shape[4]
            )
            out_corr = out_corr / x1.shape[1]
            out_corr_relu = self.leakyRELU(out_corr)

            # flow estimator
            if l == 0:
                x_intm, flow = self.flow_estimators[l](out_corr_relu)
            else:
                x_intm, flow = self.flow_estimators[l](
                    torch.cat([out_corr_relu, x1, flow], dim=1)
                )

            # upsampling or post-processing
            if l != self.output_level:
                flows.append(flow)
            else:
                flow_res = self.context_networks(torch.cat([x_intm, flow], dim=1))
                flow = flow + flow_res
                flows.append(flow)
                break

        flow_up = upsample2d_as(flow, x1_raw, mode="bilinear") * (1.0 / self.div_flow)
        flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)

        outputs = {}
        if self.training:
            outputs["flow_preds"] = flows
            outputs["flows"] = flow_up[:, None]
        else:
            outputs["flows"] = flow_up[:, None]
        return outputs


@register_model
@trainable
class irr_pwcnet(IRRPWCNet):
    pass
